"""Kubernetes adaptors

Thread safety notes:

The API functions (core_api, batch_api, etc.) return cached clients that are
created with context-specific ApiClient instances.

Set SKYPILOT_KUBECONFIG_REFRESH_INTERVAL_SECONDS (seconds) to refresh the
client proactively at a fixed interval so it is rebuilt from the updated
kubeconfig (e.g. for short-lived certs).
"""
import functools
import logging
import os
import platform
import threading
import time
import typing
from typing import Any, Callable, Optional, Set

from sky import sky_logging
from sky.adaptors import common
from sky.utils import annotations
from sky.utils import common_utils
from sky.utils import ux_utils

if typing.TYPE_CHECKING:
    import kubernetes
    import urllib3
    import urllib3.exceptions
else:
    _IMPORT_ERROR_MESSAGE = ('Failed to import dependencies for Kubernetes. '
                             'Try running: pip install "skypilot[kubernetes]"')
    kubernetes = common.LazyImport('kubernetes',
                                   import_error_message=_IMPORT_ERROR_MESSAGE)
    urllib3 = common.LazyImport('urllib3',
                                import_error_message=_IMPORT_ERROR_MESSAGE)

# Timeout to use for API calls
API_TIMEOUT = 5

# Check if KUBECONFIG is set, and use it if it is.
DEFAULT_KUBECONFIG_PATH = '~/.kube/config'
# From kubernetes package, keep a copy here to avoid actually importing
# kubernetes package when parsing the KUBECONFIG env var to do credential
# file mounts.
ENV_KUBECONFIG_PATH_SEPARATOR = ';' if platform.system() == 'Windows' else ':'

DEFAULT_IN_CLUSTER_REGION = 'in-cluster'
# The name for the environment variable that stores the in-cluster context name
# for Kubernetes clusters. This is used to associate a name with the current
# context when running with in-cluster auth. If not set, the context name is
# set to DEFAULT_IN_CLUSTER_REGION.
IN_CLUSTER_CONTEXT_NAME_ENV_VAR = 'SKYPILOT_IN_CLUSTER_CONTEXT_NAME'
# If set (positive seconds), client is refreshed proactively after this
# interval.
KUBECONFIG_REFRESH_INTERVAL_ENV_VAR = (
    'SKYPILOT_KUBECONFIG_REFRESH_INTERVAL_SECONDS')

logger = sky_logging.init_logger(__name__)


def _decorate_methods(obj: Any, decorator: Callable, decoration_type: str):
    for attr_name in dir(obj):
        attr = getattr(obj, attr_name)
        # Skip methods starting with '__' since they are invoked through one
        # of the main methods, which are already decorated.
        if callable(attr) and not attr_name.startswith('__'):
            decorated_types: Set[str] = getattr(attr, '_sky_decorator_types',
                                                set())
            if decoration_type not in decorated_types:
                decorated_attr = decorator(attr)
                decorated_attr._sky_decorator_types = (  # pylint: disable=protected-access
                    decorated_types | {decoration_type})
                setattr(obj, attr_name, decorated_attr)
    return obj


def _api_logging_decorator(logger_src: str, level: int):
    """Decorator to set logging level for API calls.

    This is used to suppress the verbose logging from urllib3 when calls to the
    Kubernetes API timeout.
    """

    def decorated_api(api):

        def wrapped(*args, **kwargs):
            obj = api(*args, **kwargs)
            _decorate_methods(obj,
                              sky_logging.set_logging_level(logger_src, level),
                              'api_log')
            return obj

        return wrapped

    return decorated_api


def _get_config_file() -> str:
    # Kubernetes load the kubeconfig from the KUBECONFIG env var on
    # package initialization. So we have to reload the KUBECONFIG env var
    # everytime in case the KUBECONFIG env var is changed.
    return os.environ.get('KUBECONFIG', DEFAULT_KUBECONFIG_PATH)


def _get_api_client(context: Optional[str] = None) -> Any:
    """Get an ApiClient for the given context without modifying global config.

    This is fully thread-safe because it creates isolated Configuration
    objects for each client rather than modifying the global
    kubernetes.client.configuration.

    Args:
        context: The Kubernetes context to use. If None, tries in-cluster config
            first, then falls back to kubeconfig current-context.

    Returns:
        A kubernetes.client.ApiClient configured for the specified context.

    Raises:
        ValueError: If the configuration cannot be loaded.
    """
    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

    def _get_api_client_from_kubeconfig(context: Optional[str] = None) -> Any:
        """Load kubeconfig, return ApiClient without modifying global state."""
        try:
            # new_client_from_config returns an ApiClient configured for the
            # specified context WITHOUT modifying the global configuration.
            # This is the key to thread-safety.
            return kubernetes.config.new_client_from_config(
                config_file=_get_config_file(), context=context)
        except kubernetes.config.config_exception.ConfigException as e:
            suffix = common_utils.format_exception(e, use_bracket=True)
            context_name = '(current-context)' if context is None else context
            is_ssh_node_pool = False
            if context_name.startswith('ssh-'):
                context_name = common_utils.removeprefix(context_name, 'ssh-')
                is_ssh_node_pool = True
            # Check if exception was due to no current-context
            if 'Expected key current-context' in str(e):
                if is_ssh_node_pool:
                    context_name = common_utils.removeprefix(
                        context_name, 'ssh-')
                    err_str = ('Failed to load SSH Node Pool configuration for '
                               f'{context_name!r}.\n'
                               '    Run `sky ssh up --infra {context_name}` to '
                               'set up or repair the cluster.')
                else:
                    err_str = (
                        'Failed to load Kubernetes configuration for '
                        f'{context_name!r}. '
                        'Kubeconfig does not contain any valid context(s).'
                        f'\n{suffix}\n'
                        '    If you were running a local Kubernetes '
                        'cluster, run `sky local up` to start the cluster.')
            else:
                kubeconfig_path = os.environ.get('KUBECONFIG', '~/.kube/config')
                if is_ssh_node_pool:
                    err_str = (
                        f'Failed to load SSH Node Pool configuration for '
                        f'{context_name!r}. Run `sky ssh up --infra '
                        f'{context_name}` to set up or repair the cluster.')
                else:
                    err_str = (
                        'Failed to load Kubernetes configuration for '
                        f'{context_name!r}. Please check if your kubeconfig '
                        f'file exists at {kubeconfig_path} and is valid.'
                        f'\n{suffix}\n')
            if is_ssh_node_pool:
                err_str += (f'\nTo disable SSH Node Pool {context_name!r}: '
                            'run `sky check`.')
            else:
                err_str += (
                    '\nHint: Kubernetes attempted to query the current-context '
                    'set in kubeconfig. Check if the current-context is valid.')
            with ux_utils.print_exception_no_traceback():
                raise ValueError(err_str) from None

    if context == in_cluster_context_name() or context is None:
        try:
            # Load in-cluster config if running in a pod and context is None.
            # Use InClusterConfigLoader with an explicit Configuration object
            # to avoid modifying global state (thread-safe).
            #
            # Workaround: Kubernetes service discovery environment variables
            # may not show up in SkyPilot tasks. We set them to DNS names as
            # a fallback. See: github.com/skypilot-org/skypilot/issues/2287
            if 'KUBERNETES_SERVICE_HOST' not in os.environ:
                os.environ['KUBERNETES_SERVICE_HOST'] = 'kubernetes.default.svc'
            if 'KUBERNETES_SERVICE_PORT' not in os.environ:
                os.environ['KUBERNETES_SERVICE_PORT'] = '443'

            config = kubernetes.client.Configuration()
            kubernetes.config.load_incluster_config(config)
            return kubernetes.client.ApiClient(configuration=config)
        except kubernetes.config.config_exception.ConfigException:
            if context == in_cluster_context_name():
                # Explicitly requested in-cluster context but not in a cluster
                raise
            # Otherwise, if context is None, fall through to kubeconfig

    return _get_api_client_from_kubeconfig(context)


def list_kube_config_contexts():
    return kubernetes.config.list_kube_config_contexts(_get_config_file())


@functools.lru_cache(maxsize=None)
def _get_kubeconfig_refresh_interval_seconds() -> float:
    """Parse refresh interval from env; 0 means disabled.

    Result is cached because this is called on every k8s API method invocation
    and the env var is not expected to change at runtime.
    """
    raw = os.environ.get(KUBECONFIG_REFRESH_INTERVAL_ENV_VAR, '').strip()
    if not raw:
        return 0.0
    try:
        val = float(raw)
        return max(0.0, val)
    except ValueError:
        logger.warning(
            f'Invalid value for {KUBECONFIG_REFRESH_INTERVAL_ENV_VAR}: '
            f'"{raw}". Expected a numeric value. Disabling client '
            'refresh interval.')
        return 0.0


class RetryableClientWrapper:
    """Wrap a kubernetes client for interval-based refresh and resource cleanup.

    Each wrapper tracks its own last-refresh time and refreshes only its
    underlying client when the configured interval has elapsed, without
    invalidating other wrappers or global caches. Closes the underlying
    ApiClient on GC to release external resources (e.g. semaphores).
    """

    def __init__(self, client: Any, getter: Callable, getter_args: tuple,
                 getter_kwargs: dict):
        self._client = client
        self._getter = getter
        self._getter_args = getter_args
        self._getter_kwargs = getter_kwargs
        self._last_refresh_time = time.time()
        self._refresh_lock = threading.Lock()

    def _should_refresh(self) -> bool:
        """True if this wrapper's refresh interval has elapsed."""
        interval = _get_kubeconfig_refresh_interval_seconds()
        if interval <= 0:
            return False
        return (time.time() - self._last_refresh_time) >= interval

    def _close_client(self, client: Any) -> None:
        """Close the underlying ApiClient to release external resources."""
        try:
            real_client = None
            if isinstance(client, kubernetes.client.ApiClient):
                real_client = client
            elif isinstance(client, kubernetes.watch.Watch):
                real_client = getattr(client, '_api_client', None)
            else:
                # Typed clients (CoreV1Api etc.) are codegen'd and all have
                # an 'api_client' attribute pointing to the real ApiClient.
                real_client = getattr(client, 'api_client', None)
            if real_client is not None:
                real_client.close()
            else:
                # logger may already be cleaned up during __del__ at shutdown
                if logger is not None:
                    logger.debug(f'No client found for {client}')
        except Exception as e:  # pylint: disable=broad-except
            if logger is not None:
                logger.debug(f'Error closing Kubernetes client: {e}')

    def __getattr__(self, name: str) -> Any:
        attr = getattr(self._client, name)
        if not callable(attr):
            return attr

        @functools.wraps(attr)
        def with_refresh(*args, **kwargs):
            if self._should_refresh():
                with self._refresh_lock:
                    if self._should_refresh():
                        logger.debug(
                            'Refreshing Kubernetes client from kubeconfig '
                            'due to interval expiry.')
                        old_client = self._client
                        self._client = self._getter(*self._getter_args,
                                                    **self._getter_kwargs)
                        self._last_refresh_time = time.time()
                        self._close_client(old_client)
            method = getattr(self._client, name)
            return method(*args, **kwargs)

        # Cache on the instance so repeated accesses to the same method name
        # return the same closure without going through __getattr__ again.
        # The closure always reads self._client at call time, so it stays
        # correct after a client refresh.
        self.__dict__[name] = with_refresh
        return with_refresh

    def __del__(self):
        self._close_client(self._client)


def _retryable_kubernetes_client(getter: Callable) -> Callable:
    """Wrap a kubernetes client getter in a RetryableClientWrapper.

    On each call the getter is invoked to obtain the raw client, which is then
    wrapped so it can be transparently refreshed when the configured kubeconfig
    refresh interval has elapsed.
    """

    @functools.wraps(getter)
    def wrapper(*args: Any, **kwargs: Any) -> RetryableClientWrapper:
        client = getter(*args, **kwargs)
        return RetryableClientWrapper(client, getter, args, kwargs)

    return wrapper


@_api_logging_decorator('urllib3', logging.ERROR)
@annotations.lru_cache(scope='request')
@_retryable_kubernetes_client
def core_api(context: Optional[str] = None):
    return kubernetes.client.CoreV1Api(api_client=_get_api_client(context))


@_api_logging_decorator('urllib3', logging.ERROR)
@annotations.lru_cache(scope='request')
@_retryable_kubernetes_client
def storage_api(context: Optional[str] = None):
    return kubernetes.client.StorageV1Api(api_client=_get_api_client(context))


@_api_logging_decorator('urllib3', logging.ERROR)
@annotations.lru_cache(scope='request')
@_retryable_kubernetes_client
def auth_api(context: Optional[str] = None):
    return kubernetes.client.RbacAuthorizationV1Api(
        api_client=_get_api_client(context))


@_api_logging_decorator('urllib3', logging.ERROR)
@annotations.lru_cache(scope='request')
@_retryable_kubernetes_client
def networking_api(context: Optional[str] = None):
    return kubernetes.client.NetworkingV1Api(
        api_client=_get_api_client(context))


@_api_logging_decorator('urllib3', logging.ERROR)
@annotations.lru_cache(scope='request')
@_retryable_kubernetes_client
def custom_objects_api(context: Optional[str] = None):
    return kubernetes.client.CustomObjectsApi(
        api_client=_get_api_client(context))


@_api_logging_decorator('urllib3', logging.ERROR)
@annotations.lru_cache(scope='request')
@_retryable_kubernetes_client
def node_api(context: Optional[str] = None):
    return kubernetes.client.NodeV1Api(api_client=_get_api_client(context))


@_api_logging_decorator('urllib3', logging.ERROR)
@annotations.lru_cache(scope='request')
@_retryable_kubernetes_client
def apps_api(context: Optional[str] = None):
    return kubernetes.client.AppsV1Api(api_client=_get_api_client(context))


@_api_logging_decorator('urllib3', logging.ERROR)
@annotations.lru_cache(scope='request')
@_retryable_kubernetes_client
def batch_api(context: Optional[str] = None):
    return kubernetes.client.BatchV1Api(api_client=_get_api_client(context))


@_api_logging_decorator('urllib3', logging.ERROR)
@annotations.lru_cache(scope='request')
@_retryable_kubernetes_client
def api_client(context: Optional[str] = None):
    return _get_api_client(context)


@_api_logging_decorator('urllib3', logging.ERROR)
@annotations.lru_cache(scope='request')
@_retryable_kubernetes_client
def custom_resources_api(context: Optional[str] = None):
    return kubernetes.client.CustomObjectsApi(
        api_client=_get_api_client(context))


@_api_logging_decorator('urllib3', logging.ERROR)
@annotations.lru_cache(scope='request')
@_retryable_kubernetes_client
def watch(context: Optional[str] = None):
    w = kubernetes.watch.Watch()
    w._api_client = _get_api_client(context)  # pylint: disable=protected-access
    return w


def api_exception():
    return kubernetes.client.rest.ApiException


def config_exception():
    return kubernetes.config.config_exception.ConfigException


def max_retry_error():
    return urllib3.exceptions.MaxRetryError


def stream():
    return kubernetes.stream.stream


def in_cluster_context_name() -> Optional[str]:
    """Returns the name of the in-cluster context from the environment.

    If the environment variable is not set, returns the default in-cluster
    context name.
    """
    return (os.environ.get(IN_CLUSTER_CONTEXT_NAME_ENV_VAR) or
            DEFAULT_IN_CLUSTER_REGION)



"""Initialize docker containers on a remote node."""

import dataclasses
import re
import shlex
import time
from typing import Any, Dict, List, Optional

from sky import sky_logging
from sky.skylet import constants
from sky.utils import command_runner
from sky.utils import subprocess_utils

logger = sky_logging.init_logger(__name__)

# Configure environment variables. A docker image can have environment variables
# set in the Dockerfile with `ENV``. We need to export these variables to the
# shell environment, so that our ssh session can access them.
# Filter out RAY_RUNTIME_ENV_HOOK to prevent Ray version conflicts.
# Docker images with Ray 2.48.0+ set this for UV package manager support,
# but it causes FAILED_DRIVER errors with SkyPilot's Ray 2.9.3.
# See: https://github.com/skypilot-org/skypilot/pull/7181
SETUP_ENV_VARS_CMD = (
    'prefix_cmd() '
    '{ if [ $(id -u) -ne 0 ]; then echo "sudo"; else echo ""; fi; } && '
    'export -p | grep -v RAY_RUNTIME_ENV_HOOK > ~/container_env_var.sh && '
    '$(prefix_cmd) '
    'mv ~/container_env_var.sh /etc/profile.d/container_env_var.sh;')

# Docker daemon may not be ready when the machine is firstly started. The error
# message starts with the following string. We should wait for a while and retry
# the command.
DOCKER_PERMISSION_DENIED_STR = ('permission denied while trying to connect to ')

DOCKER_SOCKET_NOT_READY_STR = ('Is the docker daemon running?')
DOCKER_SOCKET_NOT_READY_STR_2 = (
    'check if the path is correct and if the daemon is running')

_DOCKER_SOCKET_WAIT_TIMEOUT_SECONDS = 30

# Install AWS CLI v2 (not v1 from pip) as it's required for ECR authentication
# AWS CLI v2 is installed as a standalone binary, not a Python package. See:
# https://docs.aws.amazon.com/cli/latest/userguide/getting-started-install.html
INSTALL_AWS_CLI_CMD = (
    'which aws || ((command -v unzip >/dev/null 2>&1 || '
    '(sudo apt-get update && sudo apt-get install -y unzip)) && '
    'curl -fsSL "https://awscli.amazonaws.com/awscli-exe-linux-x86_64.zip" '
    '-o "/tmp/awscliv2.zip" && '
    'unzip -q /tmp/awscliv2.zip -d /tmp && sudo /tmp/aws/install '
    '&& rm -rf /tmp/awscliv2.zip /tmp/aws)')

# Pattern to extract SSH user from command output, handling MOTD contamination
_DOCKER_USER_PATTERN = re.compile(r'SKYPILOT_DOCKER_USER: ([^\s\n]+)')


def _extract_region_from_ecr_server(server: str) -> str:
    """Extract AWS region from ECR server URL.

    ECR server format: <account-id>.dkr.ecr.<region>.amazonaws.com
    Returns the region part from the URL.
    """
    # Split: ['<account-id>', 'dkr', 'ecr', '<region>', 'amazonaws', 'com']
    parts = server.split('.')
    if len(parts) >= 6 and parts[1] == 'dkr' and parts[2] == 'ecr':
        return parts[3]
    raise ValueError(f'Invalid ECR server format: {server}')


@dataclasses.dataclass
class DockerLoginConfig:
    """Config for docker login. Used for pulling from private registries."""
    username: str
    password: str
    server: str

    def format_image(self, image: str) -> str:
        """Format the image name with the server prefix."""
        server_prefix = f'{self.server}/'
        if not image.startswith(server_prefix):
            return f'{server_prefix}{image}'
        return image

    @classmethod
    def from_env_vars(cls, d: Dict[str, str]) -> 'DockerLoginConfig':
        return cls(
            username=d[constants.DOCKER_USERNAME_ENV_VAR],
            password=d[constants.DOCKER_PASSWORD_ENV_VAR],
            server=d[constants.DOCKER_SERVER_ENV_VAR],
        )


# Copied from ray.autoscaler._private.ray_constants
# The default maximum number of bytes to allocate to the object store unless
# overridden by the user.
DEFAULT_OBJECT_STORE_MAX_MEMORY_BYTES = 200 * 10**9
# The default proportion of available memory allocated to the object store
DEFAULT_OBJECT_STORE_MEMORY_PROPORTION = 0.3


def _check_helper(cname, template, docker_cmd):
    return ' '.join([
        docker_cmd, 'inspect', '-f', '"{{' + template + '}}"', cname, '||',
        'true'
    ])


def check_docker_running_cmd(cname, docker_cmd):
    return _check_helper(cname, '.State.Running', docker_cmd)


def check_bind_mounts_cmd(cname, docker_cmd):
    return _check_helper(cname, 'json .Mounts', docker_cmd)


def check_docker_image(cname, docker_cmd):
    return _check_helper(cname, '.Config.Image', docker_cmd)


def maybe_remove_container_cmds(container_name, docker_cmd):
    """Remove the container if it exists. If not, it will be a no-op.
    """
    docker_rm = [
        docker_cmd,
        'rm',
        '-f',
        container_name,
        '2>/dev/null',
        '||',
        'true',
    ]
    return ' '.join(docker_rm)


def docker_start_cmds(
    image,
    container_name,
    user_options,
    docker_cmd,
):
    """Generating docker start command.

    The code is borrowed from `ray.autoscaler._private.command_runner`.
    We made the following two changes:
      1. Remove `--rm` to keep the container after `ray stop` is executed.
      2. Remove mount options, as all the file mounts will be handled after
        the container is started, through `rsync` command.
    """

    # for click, used in ray cli
    env_vars = {'LC_ALL': 'C.UTF-8', 'LANG': 'C.UTF-8'}
    env_flags = ' '.join(
        ['-e {name}={val}'.format(name=k, val=v) for k, v in env_vars.items()])

    user_options_str = ' '.join(user_options)
    docker_run = [
        docker_cmd,
        'run',
        # SkyPilot: Remove --rm flag to keep the container after `ray stop`
        # is executed.
        '--name {}'.format(container_name),
        '-d',
        '-it',
        env_flags,
        user_options_str,
        '--net=host',
        # SkyPilot: Add following options to enable fuse and resource limits.
        '--cap-add=SYS_ADMIN',
        '--cap-add=SYS_RESOURCE',
        '--device=/dev/fuse',
        '--security-opt=apparmor:unconfined',
        '--entrypoint=/bin/bash',
        image,
    ]
    return ' '.join(docker_run)


def _with_interactive(cmd):
    force_interactive = (
        f'source ~/.bashrc; '
        f'export OMP_NUM_THREADS=1 PYTHONWARNINGS=ignore && ({cmd})')
    return ['bash', '--login', '-c', '-i', shlex.quote(force_interactive)]


def _redact_docker_password(cmd: str) -> str:
    parts = shlex.split(cmd)
    for i, part in enumerate(parts):
        if part.startswith('--password'):
            if part.startswith('--password='):
                parts[i] = '--password=<redacted>'
            elif i + 1 < len(parts):
                parts[i + 1] = '<redacted>'
    return ' '.join(parts)


# SkyPilot: New class to initialize docker containers on a remote node.
# Adopted from ray.autoscaler._private.command_runner.DockerCommandRunner.
class DockerInitializer:
    """Initializer for docker containers on a remote node."""

    def __init__(self, docker_config: Dict[str, Any],
                 runner: 'command_runner.CommandRunner', log_path: str):
        self.docker_config = docker_config
        self.container_name = docker_config['container_name']
        self.runner = runner
        self.home_dir: Optional[str] = None
        self.initialized = False
        # podman is not fully tested yet.
        use_podman = docker_config.get('use_podman', False)
        self.docker_cmd_name = 'podman' if use_podman else 'docker'
        # Use sudo to avoid 'permission denied' errors when the SSH user
        # is not in the docker group (e.g., on Nebius). The provisioned VM
        # always has passwordless sudo. See #8764.
        self.docker_cmd = f'sudo {self.docker_cmd_name}'
        self.log_path = log_path

    def _run(
        self,
        cmd,
        run_env='host',
        wait_for_docker_daemon: bool = False,
        separate_stderr: bool = False,
        log_err_when_fail: bool = True,
        flock_name: Optional[str] = None,
        flock_args: Optional[str] = None,
    ) -> str:

        if run_env == 'docker':
            cmd = self._docker_expand_user(cmd, any_char=True)
            cmd = ' '.join(_with_interactive(cmd))
            # SkyPilot: We do not include `-it` flag here, as that will cause
            # an error: `the input device is not a TTY`, and it works without
            # `-it` flag.
            # TODO(zhwu): ray use the `-it` flag, we need to check why.
            cmd = (f'{self.docker_cmd} exec -u 0 {self.container_name}'
                   f' /bin/bash -c {shlex.quote(cmd)} ')

        if flock_name is not None:
            flock_args = flock_args or ''
            cmd = (f'flock {flock_args} /tmp/{flock_name} '
                   f'-c {shlex.quote(cmd)}')

        # Redact the password in the login command.
        redacted_cmd = _redact_docker_password(cmd)
        logger.debug(f'+ {redacted_cmd}')
        start = time.time()
        while True:
            rc, stdout, stderr = self.runner.run(
                cmd,
                require_outputs=True,
                stream_logs=False,
                separate_stderr=separate_stderr,
                log_path=self.log_path)
            if (DOCKER_PERMISSION_DENIED_STR in stdout + stderr or
                    DOCKER_SOCKET_NOT_READY_STR in stdout + stderr or
                    DOCKER_SOCKET_NOT_READY_STR_2 in stdout + stderr):
                if wait_for_docker_daemon:
                    if time.time(
                    ) - start > _DOCKER_SOCKET_WAIT_TIMEOUT_SECONDS:
                        if rc == 0:
                            # Set returncode to 1 if failed to connect to docker
                            # daemon after timeout.
                            rc = 1
                        break
                    # Close the cached connection to make the permission update
                    # of ssh user take effect, e.g. usermod -aG docker $USER,
                    # called by cloud-init of Azure.
                    self.runner.close_cached_connection()
                    logger.info(
                        'Failed to connect to docker daemon. It might be '
                        'initializing, retrying in 5 seconds...')
                    time.sleep(5)
                    continue
            break
        subprocess_utils.handle_returncode(
            rc,
            redacted_cmd,
            error_msg='Failed to run docker setup commands.',
            stderr=stdout + stderr,
            # Print out the error message if the command failed.
            stream_logs=log_err_when_fail)
        return stdout.strip()

    def initialize(self) -> str:
        specific_image = self.docker_config['image']

        self._check_docker_installed()

        # SkyPilot: Check if the container is exited but not removed.
        # If true, then we can start the container directly.
        # Notice that we will skip all setup commands, so we need to
        # manually start the ssh service.
        if self._check_container_exited():
            self.initialized = True
            self._run(f'{self.docker_cmd} start {self.container_name}')
            self._run('sudo service ssh start',
                      run_env='docker',
                      flock_name=f'{self.container_name}.sky.lifecycle.lock',
                      flock_args='-s -w 1')
            return self._run('whoami', run_env='docker')

        # SkyPilot: Docker login if user specified a private docker registry.
        if 'docker_login_config' in self.docker_config:
            docker_login_config = DockerLoginConfig(
                **self.docker_config['docker_login_config'])

            if docker_login_config.password:
                # Password is allowed to be empty, in that case, we will not run
                # the login command, and assume that the image pulling is
                # authenticated by the IAM permission on the VM.
                self._run(
                    f'{self.docker_cmd} login --username '
                    f'{shlex.quote(docker_login_config.username)} '
                    f'--password {shlex.quote(docker_login_config.password)} '
                    f'{shlex.quote(docker_login_config.server)}',
                    wait_for_docker_daemon=True)
            elif (docker_login_config.server.endswith('.amazonaws.com') and
                  '.dkr.ecr.' in docker_login_config.server):
                # AWS ECR: Use aws ecr get-login-password for authentication
                # ECR format: <account-id>.dkr.ecr.<region>.amazonaws.com
                # This command uses the IAM credentials from the EC2 instance
                # Ref: https://docs.aws.amazon.com/AmazonECR/latest/userguide/registry_auth.html # pylint: disable=line-too-long
                region = _extract_region_from_ecr_server(
                    docker_login_config.server)

                # AWS CLI is not pre-installed on AWS instances, unlike gcloud
                # on GCP instances, so we need to install it first
                self._run(INSTALL_AWS_CLI_CMD, wait_for_docker_daemon=False)

                self._run(
                    f'aws ecr get-login-password --region {region} | '
                    f'{self.docker_cmd} login --username AWS '
                    f'--password-stdin '
                    f'{shlex.quote(docker_login_config.server)}',
                    wait_for_docker_daemon=True)
            elif docker_login_config.server.endswith('-docker.pkg.dev'):
                # Docker image server is on GCR, we need to do additional setup
                # to pull the image.
                # When no username or password is provided, we assume that
                # we are on GCP VM (i.e. gcloud auth configure-docker is
                # enough), or the image server is public.
                # For the former case, gcloud should be available, and latter
                # should be fine to fail the following command.
                # Use sudo so the credential helper is configured in root's
                # docker config (/root/.docker/config.json), since docker
                # commands run with sudo. See #8906.
                self._run('sudo gcloud auth configure-docker '
                          f'{shlex.quote(docker_login_config.server)} '
                          '--quiet || true')
            # We automatically add the server prefix to the image name if
            # the user did not add it.
            specific_image = docker_login_config.format_image(specific_image)

        if self.docker_config.get('pull_before_run', True):
            assert specific_image, ('Image must be included in config if ' +
                                    'pull_before_run is specified')
            self._run(f'{self.docker_cmd} pull {specific_image}',
                      wait_for_docker_daemon=True)
        else:
            self._run(
                f'{self.docker_cmd} image inspect {specific_image} '
                '1> /dev/null  2>&1 || '
                f'{self.docker_cmd} pull {specific_image}',
                wait_for_docker_daemon=True)

        logger.info(f'Starting container {self.container_name} with image '
                    f'{specific_image}')
        container_running = self._check_container_status()
        if container_running:
            running_image = self._run(
                check_docker_image(self.container_name, self.docker_cmd))
            if running_image != specific_image:
                logger.error(
                    f'A container with name {self.container_name} is running '
                    f'image {running_image} instead of {specific_image} (which '
                    'was provided in the YAML)')
        else:
            # Edit docker config first to avoid disconnecting the container
            # from GPUs when a systemctl command is called. This is a known
            # issue with nvidia container toolkit:
            # https://github.com/NVIDIA/nvidia-container-toolkit/issues/48
            self._run(
                '{ which jq || sudo apt update && sudo apt install -y jq; } && '
                '{ [ -f /etc/docker/daemon.json ] || '
                'echo "{}" | sudo tee /etc/docker/daemon.json;'
                'sudo jq \'.["exec-opts"] = ["native.cgroupdriver=cgroupfs"]\' '
                '/etc/docker/daemon.json > /tmp/daemon.json;'
                'sudo mv /tmp/daemon.json /etc/docker/daemon.json;'
                'sudo systemctl restart docker; } || true')
            user_docker_run_options = self.docker_config.get('run_options', [])
            remove_container_cmd = maybe_remove_container_cmds(
                self.container_name,
                self.docker_cmd,
            )
            start_command = docker_start_cmds(
                specific_image,
                self.container_name,
                self._configure_runtime(
                    self._auto_configure_shm(user_docker_run_options)),
                self.docker_cmd,
            )
            self._run(f'{remove_container_cmd} && {start_command}',
                      flock_name=f'{self.container_name}.sky.lifecycle.lock',
                      flock_args='-x -w 10')

        # SkyPilot: Setup Commands.
        # TODO(zhwu): the following setups should be aligned with the kubernetes
        # pod setup, like provision.kubernetes.instance::_set_env_vars_in_pods
        # TODO(tian): These setup commands assumed that the container is
        # debian-based. We should make it more general.
        # Most of docker images are using root as default user, so we set an
        # alias for sudo to empty string, therefore any sudo in the following
        # commands won't fail.
        # Disable apt-get from asking user input during installation.
        # see https://askubuntu.com/questions/909277/avoiding-user-interaction-with-tzdata-when-installing-certbot-in-a-docker-contai  # pylint: disable=line-too-long
        self._run(
            f'echo \'{command_runner.ALIAS_SUDO_TO_EMPTY_FOR_ROOT_CMD}\' '
            '>> ~/.bashrc;'
            'echo "export DEBIAN_FRONTEND=noninteractive" >> ~/.bashrc;',
            run_env='docker')
        # Install dependencies.
        cmd = (
            'bash -lc \''
            'exec 200>/var/tmp/sky_apt.lock; '
            'flock -x -w 120 200 || exit 1; '
            'export DEBIAN_FRONTEND=noninteractive; '
            'apt-get -yq update && '
            # Our mount script will install gcsfuse without fuse package.
            # We need to install fuse package first to enable storage mount.
            # The dpkg option is to suppress the prompt for fuse installation.
            'apt-get -o DPkg::Options::=--force-confnew install -y '
            'rsync curl wget patch openssh-server python3-pip fuse\'')
        self._run(cmd, run_env='docker')

        # Copy local authorized_keys to docker container.
        # Stop and disable jupyter service. This is to avoid port conflict on
        # 8080 if we use default deep learning image in GCP, and 8888 if we use
        # default deep learning image in Azure.
        # Azure also has a jupyterhub service running on 8081, so we stop and
        # disable that too.
        container_name = constants.DEFAULT_DOCKER_CONTAINER_NAME
        self._run(
            f'rsync -e "{self.docker_cmd} exec -i" -avz ~/.ssh/authorized_keys '
            f'{container_name}:/tmp/host_ssh_authorized_keys;'
            'sudo systemctl stop jupyter > /dev/null 2>&1 || true;'
            'sudo systemctl disable jupyter > /dev/null 2>&1 || true;'
            'sudo systemctl stop jupyterhub > /dev/null 2>&1 || true;'
            'sudo systemctl disable jupyterhub > /dev/null 2>&1 || true;',
            run_env='host')

        # Detect the container's default user for SSH. We run without -u 0
        # to get the actual default user not root. This ensures SSH connects
        # as the correct user whose home directory and SSH keys we've set up.
        # Pattern matching to prevent MOTD contamination and reliably
        # parse docker user. Refer to CommandRunner::_get_remote_home_dir.
        # Note: We use single quotes around the inner command and escape
        # them to prevent $(whoami) from being expanded on the host VM.
        docker_user_cmd = (f'{self.docker_cmd} exec {self.container_name} '
                           'bash -c \'echo SKYPILOT_DOCKER_USER: $(whoami)\'')
        docker_user_output = self._run(docker_user_cmd, separate_stderr=True)
        docker_user_match = _DOCKER_USER_PATTERN.search(docker_user_output)
        if docker_user_match:
            docker_user = docker_user_match.group(1)
        else:
            raise ValueError('Failed to find Docker user identifier: '
                             f'{docker_user_output}')

        # Change the default port of sshd from 22 to DEFAULT_DOCKER_PORT.
        # Append the host VM's authorized_keys to the container's authorized_keys.
        # This allows any machine that can ssh into the host VM to ssh into the
        # container.
        # Last command here is to eliminate the error
        # `mesg: ttyname failed: inappropriate ioctl for device`.
        # see https://www.educative.io/answers/error-mesg-ttyname-failed-inappropriate-ioctl-for-device  # pylint: disable=line-too-long
        port = constants.DEFAULT_DOCKER_PORT
        # In case the port is already configured in the sshd_config file
        # in some images, we delete it first and then append the new one.
        # pylint: disable=anomalous-backslash-in-string
        self._run(
            'sudo sed -i "/^Port .*/d" /etc/ssh/sshd_config;'
            f'echo "Port {port}" | sudo tee -a /etc/ssh/sshd_config > /dev/null;'
            'mkdir -p ~/.ssh;'
            'cat /tmp/host_ssh_authorized_keys >> ~/.ssh/authorized_keys;'
            # Set ownership of .ssh to the detected docker user, since these
            # commands run as root but SSH will connect as the default user
            # who needs write access to ~/.ssh
            f'chown -R {shlex.quote(docker_user)} ~/.ssh;'
            'sudo service ssh start;'
            'sudo sed -i "s/mesg n/tty -s \\&\\& mesg n/" ~/.profile;'
            f'{SETUP_ENV_VARS_CMD}',
            run_env='docker')

        # SkyPilot: End of Setup Commands.
        self.initialized = True
        return docker_user

    def _check_docker_installed(self):
        no_exist = 'NoExist'
        # SkyPilot: Add the current user to the docker group first (if needed),
        # before checking if docker is installed to avoid permission issues.
        docker_cmd = ('id -nG $USER | grep -qw docker || '
                      'sudo usermod -aG docker $USER > /dev/null 2>&1;'
                      f'command -v {self.docker_cmd_name} || echo {no_exist!r}')
        cleaned_output = self._run(docker_cmd)
        timeout = 60 * 10  # 10 minute timeout
        start = time.time()
        while no_exist in cleaned_output or 'docker' not in cleaned_output:
            if time.time() - start > timeout:
                logger.error(
                    f'{self.docker_cmd_name.capitalize()} not installed. '
                    f'Please use an image with '
                    f'{self.docker_cmd_name.capitalize()} installed.')
                return
            time.sleep(5)
            cleaned_output = self._run(docker_cmd)

    def _check_container_status(self):
        if self.initialized:
            return True
        output = (self._run(
            check_docker_running_cmd(self.container_name, self.docker_cmd)))
        # Checks for the false positive where 'true' is in the container name
        return 'true' in output.lower(
        ) and 'no such object' not in output.lower()

    def _docker_expand_user(self, string, any_char=False):
        user_pos = string.find('~')
        if user_pos > -1:
            if self.home_dir is None:
                cmd = (f'{self.docker_cmd} exec {self.container_name}'
                       ' printenv HOME')
                self.home_dir = self._run(
                    cmd,
                    separate_stderr=True,
                    flock_name=f'{self.container_name}.sky.lifecycle.lock',
                    flock_args='-s -w 1')
                # Check for unexpected newline in home directory, which can be
                # a common issue when the output is mixed with stderr.
                assert '\n' not in self.home_dir, (
                    'Unexpected newline in home directory '
                    f'({{self.home_dir}}) retrieved with {cmd}')

            if any_char:
                return string.replace('~/', self.home_dir + '/')

            elif not any_char and user_pos == 0:
                return string.replace('~', self.home_dir, 1)

        return string

    def _configure_runtime(self, run_options: List[str]) -> List[str]:
        if self.docker_config.get('disable_automatic_runtime_detection'):
            return run_options

        runtime_output = (self._run(f'{self.docker_cmd} ' +
                                    'info -f "{{.Runtimes}}"'))
        if 'nvidia-container-runtime' in runtime_output:
            try:
                self._run('nvidia-smi', log_err_when_fail=False)
                return run_options + ['--runtime=nvidia', '--gpus all']
            except Exception as e:  # pylint: disable=broad-except
                logger.debug(
                    'Nvidia Container Runtime is present in the docker image'
                    'specified, but no GPUs found on the cluster. It should '
                    'still if the cluster is expected to have no GPU.\n'
                    f'  Details for nvidia-smi: {e}')
                return run_options

        return run_options

    def _auto_configure_shm(self, run_options: List[str]) -> List[str]:
        if self.docker_config.get('disable_shm_size_detection'):
            return run_options
        for run_opt in run_options:
            if '--shm-size' in run_opt:
                logger.info('Bypassing automatic SHM-Detection because of '
                            f'`run_option`: {run_opt}')
                return run_options
        try:
            shm_output = self._run('cat /proc/meminfo || true')
            available_memory = int([
                ln for ln in shm_output.split('\n') if 'MemAvailable' in ln
            ][0].split()[1])
            available_memory_bytes = available_memory * 1024
            # Overestimate SHM size by 10%
            shm_size = min(
                (available_memory_bytes *
                 DEFAULT_OBJECT_STORE_MEMORY_PROPORTION * 1.1),
                DEFAULT_OBJECT_STORE_MAX_MEMORY_BYTES,
            )
            return run_options + [f'--shm-size="{shm_size}b"']
        except Exception as e:  # pylint: disable=broad-except
            logger.warning(
                f'Received error while trying to auto-compute SHM size {e}')
            return run_options

    # SkyPilot: New function to check whether a container is exited
    # (but not removed). This is due to previous `sky stop` command,
    # which will stop the container but not remove it.
    def _check_container_exited(self) -> bool:
        if self.initialized:
            return True
        output = self._run(check_docker_running_cmd(self.container_name,
                                                    self.docker_cmd),
                           wait_for_docker_daemon=True)
        return ('false' in output.lower() and
                'no such object' not in output.lower())



"""Executor for the requests.

We start limited number of workers for long-running requests, and
significantly more workers for short-running requests. This is to optimize the
resource usage and the latency of the requests.

* Long-running requests are those requests that can take a long time to finish
and more resources are needed, such as cluster launching, starting, job
submission, managed job submission, etc.

* Short-running requests are those requests that can be done quickly, and
require a quick response, such as status check, job status check, etc.

With more short-running workers, we can serve more short-running requests in
parallel, and reduce the latency.

The number of the workers is determined by the system resources.

See the [README.md](../README.md) for detailed architecture of the executor.
"""
import asyncio
import concurrent.futures
import contextlib
import multiprocessing
import os
import queue as queue_lib
import signal
import sys
import threading
import time
import typing
from typing import Any, Callable, Generator, List, Optional, TextIO, Tuple

import psutil
import setproctitle

from sky import exceptions
from sky import global_user_state
from sky import models
from sky import sky_logging
from sky import skypilot_config
from sky.adaptors import kubernetes as kubernetes_adaptor
from sky.metrics import utils as metrics_utils
from sky.server import common as server_common
from sky.server import config as server_config
from sky.server import constants as server_constants
from sky.server import daemons
from sky.server import metrics as metrics_lib
from sky.server import plugins
from sky.server.requests import payloads
from sky.server.requests import preconditions
from sky.server.requests import process
from sky.server.requests import request_names
from sky.server.requests import requests as api_requests
from sky.server.requests import threads
from sky.server.requests.queues import local_queue
from sky.server.requests.queues import mp_queue
from sky.skylet import constants
from sky.utils import annotations
from sky.utils import common_utils
from sky.utils import context
from sky.utils import context_utils
from sky.utils import subprocess_utils
from sky.utils import tempstore
from sky.utils import timeline
from sky.utils import yaml_utils
from sky.utils.db import db_utils
from sky.workspaces import core as workspaces_core

if typing.TYPE_CHECKING:
    import types

# pylint: disable=ungrouped-imports
if sys.version_info >= (3, 10):
    from typing import ParamSpec
else:
    from typing_extensions import ParamSpec

P = ParamSpec('P')
logger = sky_logging.init_logger(__name__)

# On macOS, the default start method for multiprocessing is 'fork', which
# can cause issues with certain types of resources, including those used in
# the QueueManager in mp_queue.py.
# The 'spawn' start method is generally more compatible across different
# platforms, including macOS.
multiprocessing.set_start_method('spawn', force=True)

# An upper limit of max threads for request execution per server process that
# unlikely to be reached to allow higher concurrency while still prevent the
# server process become overloaded.
_REQUEST_THREADS_LIMIT = 128

_REQUEST_THREAD_EXECUTOR_LOCK = threading.Lock()
# A dedicated thread pool executor for synced requests execution in coroutine to
# avoid:
# 1. blocking the event loop;
# 2. exhausting the default thread pool executor of event loop;
_REQUEST_THREAD_EXECUTOR: Optional[threads.OnDemandThreadExecutor] = None


def get_request_thread_executor() -> threads.OnDemandThreadExecutor:
    """Lazy init and return the request thread executor for current process."""
    global _REQUEST_THREAD_EXECUTOR
    if _REQUEST_THREAD_EXECUTOR is not None:
        return _REQUEST_THREAD_EXECUTOR
    with _REQUEST_THREAD_EXECUTOR_LOCK:
        if _REQUEST_THREAD_EXECUTOR is None:
            _REQUEST_THREAD_EXECUTOR = threads.OnDemandThreadExecutor(
                name='request_thread_executor',
                max_workers=_REQUEST_THREADS_LIMIT)
        return _REQUEST_THREAD_EXECUTOR


class RequestQueue:
    """The queue for the requests, either redis or multiprocessing.

    The elements in the queue are tuples of (request_id, ignore_return_value).
    """

    def __init__(self,
                 schedule_type: api_requests.ScheduleType,
                 backend: Optional[server_config.QueueBackend] = None) -> None:
        self.name = schedule_type.value
        self.backend = backend
        if backend == server_config.QueueBackend.MULTIPROCESSING:
            self.queue = mp_queue.get_queue(self.name)
        elif backend == server_config.QueueBackend.LOCAL:
            self.queue = local_queue.get_queue(self.name)
        else:
            raise RuntimeError(f'Invalid queue backend: {backend}')

    def put(self, request: Tuple[str, bool, bool]) -> None:
        """Put and request to the queue.

        Args:
            request: A tuple of request_id, ignore_return_value, and retryable.
        """
        self.queue.put(request)  # type: ignore

    def get(self) -> Optional[Tuple[str, bool, bool]]:
        """Get a request from the queue.

        It is non-blocking if the queue is empty, and returns None.

        Returns:
            A tuple of request_id, ignore_return_value, and retryable.
        """
        try:
            return self.queue.get(block=False)
        except queue_lib.Empty:
            return None

    def __len__(self) -> int:
        """Get the length of the queue."""
        return self.queue.qsize()


queue_backend = server_config.QueueBackend.MULTIPROCESSING


def executor_initializer(proc_group: str):
    setproctitle.setproctitle(f'SkyPilot:executor:{proc_group}:'
                              f'{multiprocessing.current_process().pid}')
    # Load plugins for executor process.
    plugins.load_plugins(plugins.ExtensionContext())
    # Executor never stops, unless the whole process is killed.
    threading.Thread(target=metrics_lib.process_monitor,
                     args=(f'worker:{proc_group}', threading.Event()),
                     daemon=True).start()


class RequestWorker:
    """A worker that polls requests from the queue and runs them.

    The worker can run at least `garanteed_parallelism` requests in parallel.
    If there are more resources available, it can spin up extra workers up to
    `garanteed_parallelism + burstable_parallelism`.
    """

    # The type of queue this worker works on.
    schedule_type: api_requests.ScheduleType
    # The least number of requests that this worker can run in parallel.
    garanteed_parallelism: int
    # The extra number of requests that this worker can run in parallel
    # if there are available CPU/memory resources.
    burstable_parallelism: int = 0

    def __init__(self, schedule_type: api_requests.ScheduleType,
                 config: server_config.WorkerConfig) -> None:
        self.schedule_type = schedule_type
        self.garanteed_parallelism = config.garanteed_parallelism
        self.burstable_parallelism = config.burstable_parallelism
        self.num_db_connections_per_worker = (
            config.num_db_connections_per_worker)
        self._thread: Optional[threading.Thread] = None
        self._cancel_event = threading.Event()

    def __str__(self) -> str:
        return f'Worker(schedule_type={self.schedule_type.value})'

    def run_in_background(self) -> None:
        # Thread dispatcher is sufficient for current scale, refer to
        # tests/load_tests/test_queue_dispatcher.py for more details.
        # Use daemon thread for automatic cleanup.
        thread = threading.Thread(target=self.run, daemon=True)
        thread.start()
        self._thread = thread

    def cancel(self) -> None:
        if self._thread is not None:
            self._cancel_event.set()
            self._thread.join()

    def process_request(self, executor: process.BurstableExecutor,
                        queue: RequestQueue) -> None:
        try:
            request_element = queue.get()
            if request_element is None:
                time.sleep(0.1)
                return
            request_id, ignore_return_value, _ = request_element
            request = api_requests.get_request(request_id, fields=['status'])
            assert request is not None, f'Request with ID {request_id} is None'
            if request.status == api_requests.RequestStatus.CANCELLED:
                return
            del request
            logger.info(f'[{self}] Submitting request: {request_id}')
            # Start additional process to run the request, so that it can be
            # cancelled when requested by a user.
            # TODO(zhwu): since the executor is reusing the request process,
            # multiple requests can share the same process pid, which may cause
            # issues with SkyPilot core functions if they rely on the exit of
            # the process, such as subprocess_daemon.py.
            fut = executor.submit_until_success(
                _request_execution_wrapper, request_id, ignore_return_value,
                self.num_db_connections_per_worker)
            # Decrement the free executor count when a request starts
            if metrics_utils.METRICS_ENABLED:
                if self.schedule_type == api_requests.ScheduleType.LONG:
                    metrics_utils.SKY_APISERVER_LONG_EXECUTORS.dec()
                elif self.schedule_type == api_requests.ScheduleType.SHORT:
                    metrics_utils.SKY_APISERVER_SHORT_EXECUTORS.dec()
            # Monitor the result of the request execution.
            threading.Thread(target=self.handle_task_result,
                             args=(fut, request_element),
                             daemon=True).start()

            logger.info(f'[{self}] Submitted request: {request_id}')
        except (Exception, SystemExit) as e:  # pylint: disable=broad-except
            # Catch any other exceptions to avoid crashing the worker process.
            logger.error(
                f'[{self}] Error processing request: '
                f'{request_id if "request_id" in locals() else ""} '
                f'{common_utils.format_exception(e, use_bracket=True)}')

    def handle_task_result(self, fut: concurrent.futures.Future,
                           request_element: Tuple[str, bool, bool]) -> None:
        try:
            fut.result()
        except concurrent.futures.process.BrokenProcessPool as e:
            # Happens when the worker process dies unexpectedly, e.g. OOM
            # killed.
            request_id, _, retryable = request_element
            # Ensure the request status.
            api_requests.set_request_failed(request_id, e)
            logger.error(
                f'Request {request_id} failed to get processed '
                f'{common_utils.format_exception(e, use_bracket=True)}')
            if retryable:
                # If the request is retryable and disrupted by broken
                # process pool, reschedule it immediately to get it
                # retried in the new process pool.
                queue = _get_queue(self.schedule_type)
                queue.put(request_element)
        except exceptions.ExecutionRetryableError as e:
            time.sleep(e.retry_wait_seconds)
            # Reset the request status to PENDING so it can be picked up again.
            # Assume retryable since the error is ExecutionRetryableError.
            request_id, _, _ = request_element
            with api_requests.update_request(request_id) as request_task:
                assert request_task is not None, request_id
                request_task.status = api_requests.RequestStatus.PENDING
            # Reschedule the request.
            queue = _get_queue(self.schedule_type)
            queue.put(request_element)
            logger.info(f'Rescheduled request {request_id} for retry')
        finally:
            # Increment the free executor count when a request finishes
            if metrics_utils.METRICS_ENABLED:
                if self.schedule_type == api_requests.ScheduleType.LONG:
                    metrics_utils.SKY_APISERVER_LONG_EXECUTORS.inc()
                elif self.schedule_type == api_requests.ScheduleType.SHORT:
                    metrics_utils.SKY_APISERVER_SHORT_EXECUTORS.inc()

    def run(self) -> None:
        # Handle the SIGTERM signal to abort the executor process gracefully.
        proc_group = f'{self.schedule_type.value}'
        if threading.current_thread() is threading.main_thread():
            signal.signal(signal.SIGTERM, _sigterm_handler)
            setproctitle.setproctitle(f'SkyPilot:worker:{proc_group}')
        queue = _get_queue(self.schedule_type)

        # Use concurrent.futures.ProcessPoolExecutor instead of
        # multiprocessing.Pool because the former is more efficient with the
        # support of lazy creation of worker processes.
        # We use executor instead of individual multiprocessing.Process to avoid
        # the overhead of forking a new process for each request, which can be
        # about 1s delay.
        try:
            executor = process.BurstableExecutor(
                garanteed_workers=self.garanteed_parallelism,
                burst_workers=self.burstable_parallelism,
                initializer=executor_initializer,
                initargs=(proc_group,))
            # Initialize the appropriate gauge for the number of free executors
            total_executors = (self.garanteed_parallelism +
                               self.burstable_parallelism)
            if metrics_utils.METRICS_ENABLED:
                if self.schedule_type == api_requests.ScheduleType.LONG:
                    metrics_utils.SKY_APISERVER_LONG_EXECUTORS.set(
                        total_executors)
                elif self.schedule_type == api_requests.ScheduleType.SHORT:
                    metrics_utils.SKY_APISERVER_SHORT_EXECUTORS.set(
                        total_executors)
            while not self._cancel_event.is_set():
                self.process_request(executor, queue)
        # TODO(aylei): better to distinct between KeyboardInterrupt and SIGTERM.
        except KeyboardInterrupt:
            pass
        finally:
            # In most cases, here we receive either ctrl-c in foreground
            # execution or SIGTERM on server exiting. Gracefully exit the
            # worker process and the executor.
            # TODO(aylei): worker may also be killed by system daemons like
            # OOM killer, crash the API server or recreate the worker process
            # to avoid broken state in such cases.
            logger.info(f'[{self}] Worker process interrupted')
            executor.shutdown()


@annotations.lru_cache(scope='global', maxsize=None)
def _get_queue(schedule_type: api_requests.ScheduleType) -> RequestQueue:
    return RequestQueue(schedule_type, backend=queue_backend)


@contextlib.contextmanager
def override_request_env_and_config(
        request_body: payloads.RequestBody, request_id: str,
        request_name: str) -> Generator[None, None, None]:
    """Override the environment and SkyPilot config for a request."""
    original_env = os.environ.copy()
    try:
        # Unset SKYPILOT_DEBUG by default, to avoid the value set on the API
        # server affecting client requests. If set on the client side, it will
        # be overridden by the request body.
        os.environ.pop('SKYPILOT_DEBUG', None)
        # Remove the db connection uri from client supplied env vars, as the
        # client should not set the db string on server side.
        request_body.env_vars.pop(constants.ENV_VAR_DB_CONNECTION_URI, None)
        # Remove the in-cluster context name from client supplied env vars.
        # When a client runs inside a Kubernetes pod (e.g., a managed job with
        # api_server_access), its env has SKYPILOT_IN_CLUSTER_CONTEXT_NAME set
        # pod template. If this leaks into the server's os.environ, it causes
        # the server to attempt in-cluster auth (load_incluster_config) instead
        # of using its own kubeconfig, which fails when the server is not
        # running in a Kubernetes pod.
        request_body.env_vars.pop(
            kubernetes_adaptor.IN_CLUSTER_CONTEXT_NAME_ENV_VAR, None)
        os.environ.update(request_body.env_vars)
        # Note: may be overridden by AuthProxyMiddleware.
        # TODO(zhwu): we need to make the entire request a context available to
        # the entire request execution, so that we can access info like user
        # through the execution.
        user = models.User(id=request_body.env_vars[constants.USER_ID_ENV_VAR],
                           name=request_body.env_vars[constants.USER_ENV_VAR])
        _, user = global_user_state.add_or_update_user(user, return_user=True)

        # Force color to be enabled.
        os.environ['CLICOLOR_FORCE'] = '1'
        server_common.reload_for_new_request(
            client_entrypoint=request_body.entrypoint,
            client_command=request_body.entrypoint_command,
            using_remote_api_server=request_body.using_remote_api_server,
            user=user,
            request_id=request_id)
        logger.debug(
            f'override path: {request_body.override_skypilot_config_path}')
        with skypilot_config.override_skypilot_config(
                request_body.override_skypilot_config,
                request_body.override_skypilot_config_path):
            # Skip permission check for sky.workspaces.get request
            # as it is used to determine which workspaces the user
            # has access to.
            if request_name != 'sky.workspaces.get':
                try:
                    # Reject requests that the user does not have permission
                    # to access.
                    workspaces_core.reject_request_for_unauthorized_workspace(
                        user)
                except exceptions.PermissionDeniedError as e:
                    logger.debug(
                        f'{request_id} permission denied to workspace: '
                        f'{skypilot_config.get_active_workspace()}: {e}')
                    raise e
            logger.debug(
                f'{request_id} permission granted to {request_name} request')
            yield
    finally:
        # We need to call the save_timeline() since atexit will not be
        # triggered as multiple requests can be sharing the same process.
        timeline.save_timeline()
        # Restore the original environment variables, so that a new request
        # won't be affected by the previous request, e.g. SKYPILOT_DEBUG
        # setting, etc. This is necessary as our executor is reusing the
        # same process for multiple requests.
        os.environ.clear()
        os.environ.update(original_env)


def _sigterm_handler(signum: int, frame: Optional['types.FrameType']) -> None:
    raise KeyboardInterrupt


def _request_execution_wrapper(request_id: str,
                               ignore_return_value: bool,
                               num_db_connections_per_worker: int = 0) -> None:
    """Wrapper for a request execution.

    It wraps the execution of a request to:
    1. Deserialize the request from the request database and serialize the
       return value/exception in the request database;
    2. Update the request status based on the execution result;
    3. Redirect the stdout and stderr of the execution to log file;
    4. Handle the SIGTERM signal to abort the request gracefully.
    5. Maintain the lifecycle of the temp dir used by the request.
    """
    pid = multiprocessing.current_process().pid
    proc = psutil.Process(pid)
    rss_begin = proc.memory_info().rss
    db_utils.set_max_connections(num_db_connections_per_worker)
    # Handle the SIGTERM signal to abort the request processing gracefully.
    # Only set up signal handlers in the main thread, as signal.signal() raises
    # ValueError if called from a non-main thread (e.g., in tests).
    if threading.current_thread() is threading.main_thread():
        signal.signal(signal.SIGTERM, _sigterm_handler)

    logger.info(f'Running request {request_id} with pid {pid}')

    original_stdout = original_stderr = None

    def _save_current_output() -> None:
        """Save the current stdout and stderr file descriptors."""
        nonlocal original_stdout, original_stderr
        original_stdout = os.dup(sys.stdout.fileno())
        original_stderr = os.dup(sys.stderr.fileno())

    def _redirect_output(file: TextIO) -> None:
        """Redirect stdout and stderr to the log file."""
        # Get the file descriptor from the file object
        fd = file.fileno()
        # Copy this fd to stdout and stderr
        os.dup2(fd, sys.stdout.fileno())
        os.dup2(fd, sys.stderr.fileno())

    def _restore_output() -> None:
        """Restore stdout and stderr to their original file descriptors."""
        nonlocal original_stdout, original_stderr
        if original_stdout is not None:
            os.dup2(original_stdout, sys.stdout.fileno())
            os.close(original_stdout)
            original_stdout = None

        if original_stderr is not None:
            os.dup2(original_stderr, sys.stderr.fileno())
            os.close(original_stderr)
            original_stderr = None

    request_name = None
    try:
        # As soon as the request is updated with the executor PID, we can
        # receive SIGTERM from cancellation. So, we update the request inside
        # the try block to ensure we have the KeyboardInterrupt handling.
        with api_requests.update_request(request_id) as request_task:
            assert request_task is not None, request_id
            if request_task.status != api_requests.RequestStatus.PENDING:
                logger.debug(f'Request is already {request_task.status.value}, '
                             f'skipping execution')
                return
            log_path = request_task.log_path
            request_task.pid = pid
            request_task.status = api_requests.RequestStatus.RUNNING
            func = request_task.entrypoint
            request_body = request_task.request_body
            request_name = request_task.name

        # Store copies of the original stdout and stderr file descriptors
        # We do this in two steps because we should make sure to restore the
        # original values even if we are cancelled or fail during the redirect.
        _save_current_output()

        # Append to the log file instead of overwriting it since there might be
        # logs from previous retries.
        with log_path.open('a', encoding='utf-8') as f:
            # Redirect the stdout/stderr before overriding the environment and
            # config, as there can be some logs during override that needs to be
            # captured in the log file.
            _redirect_output(f)

            # Skip debug logging for daemon requests since the daemon
            # requests has its own log level config and we don't want to
            # duplicate the daemon logs.
            debug_log_ctx = (contextlib.nullcontext()
                             if daemons.is_daemon_request_id(request_id) else
                             sky_logging.add_debug_log_handler(request_id))
            with debug_log_ctx, \
                override_request_env_and_config(
                    request_body, request_id, request_name), \
                tempstore.tempdir():
                if sky_logging.logging_enabled(logger, sky_logging.DEBUG):
                    config = skypilot_config.to_dict()
                    logger.debug(f'request config: \n'
                                 f'{yaml_utils.dump_yaml_str(dict(config))}')
                (metrics_utils.SKY_APISERVER_PROCESS_EXECUTION_START_TOTAL.
                 labels(request=request_name, pid=pid).inc())
                with metrics_utils.time_it(name=request_name,
                                           group='request_execution'):
                    return_value = func(**request_body.to_kwargs())
                f.flush()
    except KeyboardInterrupt:
        logger.info(f'Request {request_id} cancelled by user')
        # Kill all children processes related to this request.
        # Each executor handles a single request, so we can safely kill all
        # children processes related to this request.
        # This is required as python does not pass the KeyboardInterrupt to the
        # threads that are not main thread.
        subprocess_utils.kill_children_processes()
        return
    except exceptions.ExecutionRetryableError as e:
        logger.error(e)
        logger.info(e.hint)
        with api_requests.update_request(request_id) as request_task:
            assert request_task is not None, request_id
            # Retried request will undergo rescheduling and a new execution,
            # clear the pid of the request.
            request_task.pid = None
        # Yield control to the scheduler for uniform handling of retries.
        _restore_output()
        raise
    except (Exception, SystemExit) as e:  # pylint: disable=broad-except
        api_requests.set_request_failed(request_id, e)
        # Manually reset the original stdout and stderr file descriptors early
        # so that the "Request xxxx failed due to ..." log message will be
        # written to the original stdout and stderr file descriptors.
        _restore_output()
        logger.error(f'Request {request_id} failed due to '
                     f'{common_utils.format_exception(e)}')
        return
    else:
        api_requests.set_request_succeeded(
            request_id, return_value if not ignore_return_value else None)
        # Manually reset the original stdout and stderr file descriptors early
        # so that the "Request xxxx failed due to ..." log message will be
        # written to the original stdout and stderr file descriptors.
        _restore_output()
        logger.info(f'Request {request_id} finished')
    finally:
        _restore_output()
        try:
            # Capture the peak RSS before GC.
            peak_rss = max(proc.memory_info().rss, metrics_lib.peak_rss_bytes)
            # Clear request level cache to release all memory used by the
            # request.
            annotations.clear_request_level_cache()
            with metrics_utils.time_it(name='release_memory', group='internal'):
                common_utils.release_memory()
            if request_name is not None:
                _record_memory_metrics(request_name, proc, rss_begin, peak_rss)
        except Exception as e:  # pylint: disable=broad-except
            logger.error(f'Failed to record memory metrics: '
                         f'{common_utils.format_exception(e)}')


_first_request = True


def _record_memory_metrics(request_name: str, proc: psutil.Process,
                           rss_begin: int, peak_rss: int) -> None:
    """Record the memory metrics for a request."""
    # Do not record full memory delta for the first request as it
    # will loads the sky core modules and make the memory usage
    # estimation inaccurate.
    global _first_request
    if _first_request:
        _first_request = False
        return
    rss_end = proc.memory_info().rss

    # Answer "how much RSS this request contributed?"
    metrics_utils.SKY_APISERVER_REQUEST_RSS_INCR_BYTES.labels(
        name=request_name).observe(max(rss_end - rss_begin, 0))
    # Estimate the memory usage by the request by capturing the
    # peak memory delta during the request execution.
    metrics_utils.SKY_APISERVER_REQUEST_MEMORY_USAGE_BYTES.labels(
        name=request_name).observe(max(peak_rss - rss_begin, 0))


class CoroutineTask:
    """Wrapper of a background task runs in coroutine"""

    def __init__(self, task: asyncio.Task):
        self.task = task

    async def cancel(self):
        try:
            self.task.cancel()
            await self.task
        except asyncio.CancelledError:
            pass


def check_request_thread_executor_available() -> None:
    """Check if the request thread executor is available.

    This is a best effort check to hint the client to retry other server
    processes when there is no avaiable thread worker in current one. But
    a request may pass this check and still cannot get worker on execution
    time due to race condition. In this case, the client will see a failed
    request instead of retry.

    TODO(aylei): this can be refined with a refactor of our coroutine
    execution flow.
    """
    get_request_thread_executor().check_available()


def execute_request_in_coroutine(
        request: api_requests.Request) -> CoroutineTask:
    """Execute a request in current event loop.

    Args:
        request: The request to execute.

    Returns:
        A CoroutineTask handle to operate the background task.
    """
    task = asyncio.create_task(_execute_request_coroutine(request))
    return CoroutineTask(task)


def _execute_with_config_override(func: Callable,
                                  request_body: payloads.RequestBody,
                                  request_id: str, request_name: str,
                                  **kwargs) -> Any:
    """Execute a function with env and config override inside a thread."""
    # Override the environment and config within this thread's context,
    # which gets copied when we call to_thread.
    with override_request_env_and_config(request_body, request_id,
                                         request_name):
        return func(**kwargs)


async def _execute_request_coroutine(request: api_requests.Request):
    """Execute a request in current event loop.

    Similar to _request_execution_wrapper, but executed as coroutine in current
    event loop. This is designed for executing tasks that are not CPU
    intensive, e.g. sky logs.
    """
    context.initialize()
    ctx = context.get()
    assert ctx is not None, 'Context is not initialized'
    logger.info(f'Executing request {request.request_id} in coroutine')
    func = request.entrypoint
    request_body = request.request_body
    await api_requests.update_status_async(request.request_id,
                                           api_requests.RequestStatus.RUNNING)
    # Redirect stdout and stderr to the request log path.
    original_output = ctx.redirect_log(request.log_path)
    try:
        fut: asyncio.Future = context_utils.to_thread_with_executor(
            get_request_thread_executor(), _execute_with_config_override, func,
            request_body, request.request_id, request.name,
            **request_body.to_kwargs())
    except Exception as e:  # pylint: disable=broad-except
        ctx.redirect_log(original_output)
        await api_requests.set_request_failed_async(request.request_id, e)
        logger.error(f'Failed to run request {request.request_id} due to '
                     f'{common_utils.format_exception(e)}')
        return

    async def poll_task(request_id: str) -> bool:
        req_status = await api_requests.get_request_status_async(request_id)
        if req_status is None:
            raise RuntimeError('Request not found')

        if req_status.status == api_requests.RequestStatus.CANCELLED:
            ctx.cancel()
            return True

        if fut.done():
            try:
                result = await fut
                await api_requests.set_request_succeeded_async(
                    request_id, result)
            except asyncio.CancelledError:
                # The task is cancelled by ctx.cancel(), where the status
                # should already be set to CANCELLED.
                pass
            except Exception as e:  # pylint: disable=broad-except
                ctx.redirect_log(original_output)
                await api_requests.set_request_failed_async(request_id, e)
                logger.error(f'Request {request_id} failed due to '
                             f'{common_utils.format_exception(e)}')
            return True
        return False

    try:
        while True:
            res = await poll_task(request.request_id)
            if res:
                break
            await asyncio.sleep(0.5)
    except asyncio.CancelledError:
        # Current coroutine is cancelled due to client disconnect, set the
        # request status for consistency.
        await api_requests.set_request_cancelled_async(request.request_id)
        pass
    # pylint: disable=broad-except
    except (Exception, KeyboardInterrupt, SystemExit) as e:
        # Handle any other error
        ctx.redirect_log(original_output)
        await api_requests.set_request_failed_async(request.request_id, e)
        logger.error(f'Request {request.request_id} interrupted due to '
                     f'unhandled exception: {common_utils.format_exception(e)}')
        raise
    finally:
        # Always cancel the context to kill potentially running background
        # routine.
        ctx.cancel()


async def prepare_request_async(
    request_id: str,
    request_name: request_names.RequestName,
    request_body: payloads.RequestBody,
    func: Callable[P, Any],
    request_cluster_name: Optional[str] = None,
    schedule_type: api_requests.ScheduleType = (api_requests.ScheduleType.LONG),
    is_skypilot_system: bool = False,
    auth_user: Optional[models.User] = None,
) -> api_requests.Request:
    """Prepare a request for execution."""
    if auth_user is not None:
        assert auth_user.name is not None
        # Use the authenticated user identity as the single source of truth
        # if present.
        user_id = auth_user.id
        # Set user identity for executors.
        request_body.env_vars[constants.USER_ID_ENV_VAR] = user_id
        request_body.env_vars[constants.USER_ENV_VAR] = auth_user.name
    else:
        # Fallback to legacy environment variable based identity if no
        # authentication is set.
        user_id = request_body.env_vars[constants.USER_ID_ENV_VAR]
    if is_skypilot_system:
        user_id = constants.SKYPILOT_SYSTEM_USER_ID
        global_user_state.add_or_update_user(
            models.User(id=user_id,
                        name=user_id,
                        user_type=models.UserType.SYSTEM.value))
    request = api_requests.Request(
        request_id=request_id,
        name=server_constants.REQUEST_NAME_PREFIX + request_name,
        entrypoint=func,
        request_body=request_body,
        status=api_requests.RequestStatus.PENDING,
        created_at=time.time(),
        schedule_type=schedule_type,
        user_id=user_id,
        cluster_name=request_cluster_name,
        file_mounts_blob_id=getattr(request_body, 'file_mounts_blob_id', None),
    )

    if not await api_requests.create_if_not_exists_async(request):
        raise exceptions.RequestAlreadyExistsError(
            f'Request {request_id} already exists.')

    request.log_path.touch()
    return request


async def schedule_request_async(
        request_id: str,
        request_name: request_names.RequestName,
        request_body: payloads.RequestBody,
        func: Callable[P, Any],
        request_cluster_name: Optional[str] = None,
        ignore_return_value: bool = False,
        schedule_type: api_requests.ScheduleType = (
            api_requests.ScheduleType.LONG),
        is_skypilot_system: bool = False,
        precondition: Optional[preconditions.Precondition] = None,
        retryable: bool = False,
        auth_user: Optional[models.User] = None) -> None:
    """Enqueue a request to the request queue.

    Args:
        request_id: ID of the request.
        request_name: Name of the request type, e.g. "sky.launch".
        request_body: The request body containing parameters and environment
            variables.
        func: The function to execute when the request is processed.
        request_cluster_name: The name of the cluster associated with this
            request, if any.
        ignore_return_value: If True, the return value of the function will be
            ignored.
        schedule_type: The type of scheduling to use for this request, refer to
            `api_requests.ScheduleType` for more details.
        is_skypilot_system: Denote whether the request is from SkyPilot system.
        precondition: If a precondition is provided, the request will only be
            scheduled for execution when the precondition is met (returns True).
            The precondition is waited asynchronously and does not block the
            caller.
    """
    request_task = await prepare_request_async(request_id,
                                               request_name,
                                               request_body,
                                               func,
                                               request_cluster_name,
                                               schedule_type,
                                               is_skypilot_system,
                                               auth_user=auth_user)
    schedule_prepared_request(request_task, ignore_return_value, precondition,
                              retryable)


def schedule_prepared_request(request_task: api_requests.Request,
                              ignore_return_value: bool = False,
                              precondition: Optional[
                                  preconditions.Precondition] = None,
                              retryable: bool = False) -> None:
    """Enqueue a request to the request queue

    Args:
        request_task: The prepared request task to schedule.
        ignore_return_value: If True, the return value of the function will be
            ignored.
        precondition: If a precondition is provided, the request will only be
            scheduled for execution when the precondition is met (returns True).
            The precondition is waited asynchronously and does not block the
            caller.
        retryable: Whether the request should be retried if it fails.
    """

    def enqueue():
        input_tuple = (request_task.request_id, ignore_return_value, retryable)
        logger.info(f'Queuing request: {request_task.request_id}')
        _get_queue(request_task.schedule_type).put(input_tuple)

    if precondition is not None:
        # Wait async to avoid blocking caller.
        precondition.wait_async(on_condition_met=enqueue)
    else:
        enqueue()


def start(
    config: server_config.ServerConfig
) -> Tuple[Optional[multiprocessing.Process], List[RequestWorker]]:
    """Start the request workers.

    Request workers run in background, schedule the requests and delegate the
    request execution to executor processes.

    Returns:
        A tuple of the queue server process and the list of request worker
        threads.
    """
    global queue_backend
    queue_backend = config.queue_backend
    queue_server = None
    # Setup the queues.
    if queue_backend == server_config.QueueBackend.MULTIPROCESSING:
        logger.info('Creating shared request queues')
        queue_names = [
            schedule_type.value for schedule_type in api_requests.ScheduleType
        ]
        # TODO(aylei): make queue manager port configurable or pick an available
        # port automatically.
        port = mp_queue.DEFAULT_QUEUE_MANAGER_PORT
        if not common_utils.is_port_available(port):
            raise RuntimeError(
                f'SkyPilot API server fails to start as port {port!r} is '
                'already in use by another process.')
        queue_server = multiprocessing.Process(
            target=mp_queue.start_queue_manager, args=(queue_names, port))
        queue_server.start()
        mp_queue.wait_for_queues_to_be_ready(queue_names,
                                             queue_server,
                                             port=port)
    elif queue_backend == server_config.QueueBackend.LOCAL:
        # No setup is needed for local queue backend.
        pass
    else:
        # Should be checked earlier, but just in case.
        raise RuntimeError(f'Invalid queue backend: {queue_backend}')

    logger.info('Request queues created')

    workers = []
    # Start a worker for long requests.
    long_worker = RequestWorker(schedule_type=api_requests.ScheduleType.LONG,
                                config=config.long_worker_config)
    long_worker.run_in_background()
    workers.append(long_worker)

    # Start a worker for short requests.
    short_worker = RequestWorker(schedule_type=api_requests.ScheduleType.SHORT,
                                 config=config.short_worker_config)
    short_worker.run_in_background()
    workers.append(short_worker)
    return queue_server, workers



"""Add priority_class column to job_info table.

This migration adds priority_class column to store the named priority class
for managed jobs, alongside the numeric priority value.

Revision ID: 017
Revises: 016
Create Date: 2026-03-18

"""
# pylint: disable=invalid-name
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa

from sky.utils.db import db_utils

# revision identifiers, used by Alembic.
revision: str = '017'
down_revision: Union[str, Sequence[str], None] = '016'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade():
    """Add priority_class column to job_info table."""
    with op.get_context().autocommit_block():
        db_utils.add_column_to_table_alembic('job_info',
                                             'priority_class',
                                             sa.Text(),
                                             server_default=None)


def downgrade():
    """No downgrade logic."""
    pass



"""Permission service for SkyPilot API Server."""
import contextlib
import hashlib
import logging
import os
import threading
import time
from typing import Generator, List, Optional, Set

import casbin
import filelock
import sqlalchemy_adapter

from sky import global_user_state
from sky import models
from sky import sky_logging
from sky.skylet import constants
from sky.users import rbac
from sky.utils import common
from sky.utils import common_utils
from sky.utils.db import db_utils
from sky.utils.db import kv_cache

logging.getLogger('casbin.policy').setLevel(sky_logging.ERROR)
logging.getLogger('casbin.role').setLevel(sky_logging.ERROR)
logging.getLogger('casbin.model').setLevel(sky_logging.ERROR)
logging.getLogger('casbin.rbac').setLevel(sky_logging.ERROR)
logger = sky_logging.init_logger(__name__)

# Filelocks for the policy update.
POLICY_UPDATE_LOCK_PATH = os.path.expanduser('~/.sky/.policy_update.lock')
POLICY_UPDATE_LOCK_TIMEOUT_SECONDS = 20

_enforcer_instance: Optional['PermissionService'] = None

# KV cache constants for workspace permission checks.
_WORKSPACE_PERM_CACHE_PREFIX = 'perm:ws:'
_WORKSPACE_PERM_CACHE_KEY_SEP = ':'
# Long TTL as safety net; primary freshness is explicit invalidation on update.
_WORKSPACE_PERM_CACHE_TTL_SECONDS = 60 * 60  # 1h


class PermissionService:
    """Permission service for SkyPilot API Server."""

    def __init__(self):
        self.enforcer: Optional[casbin.SyncedEnforcer] = None
        self._lock = threading.Lock()

    def initialize(self):
        self._lazy_initialize(full_initialize=True)

    def _lazy_initialize(self, full_initialize: bool = False):
        if self.enforcer is not None:
            return
        with self._lock:
            if self.enforcer is not None:
                return
            global _enforcer_instance
            if _enforcer_instance is None:
                engine = global_user_state.initialize_and_get_db()
                if full_initialize:
                    db_utils.add_all_tables_to_db_sqlalchemy(
                        sqlalchemy_adapter.Base.metadata, engine)
                adapter = sqlalchemy_adapter.Adapter(
                    engine, db_class=sqlalchemy_adapter.CasbinRule)
                model_path = os.path.join(os.path.dirname(__file__),
                                          'model.conf')
                # Use SyncedEnforcer for thread safety. It uses a
                # read-write lock internally: concurrent reads (enforce,
                # get_roles_for_user) take a shared read lock, while
                # writes (load_policy, add_policy) take an exclusive
                # write lock. This prevents the RuntimeError from
                # concurrent iteration/mutation of RoleManager.all_roles.
                enforcer = casbin.SyncedEnforcer(model_path, adapter)
                self.enforcer = enforcer
                # Only set the enforcer instance once the enforcer
                # is successfully initialized, if we change it and then fail
                # we will set it to None and all subsequent calls will fail.
                _enforcer_instance = self
                if full_initialize:
                    with _policy_lock():
                        self._maybe_initialize_policies()
                        self._maybe_initialize_basic_auth_user()
            else:
                assert _enforcer_instance is not None
                self.enforcer = _enforcer_instance.enforcer

    def _ensure_enforcer(self) -> casbin.SyncedEnforcer:
        """Ensure enforcer is initialized and return it."""
        self._lazy_initialize()
        assert self.enforcer is not None, (
            'Enforcer should be initialized after _lazy_initialize()')
        return self.enforcer

    def _get_plugin_rbac_rules(self):
        """Get RBAC rules from loaded plugins.

        Returns:
            Dictionary of plugin RBAC rules, or empty dict if plugins module
            is not available or no rules are defined.
        """
        try:
            # pylint: disable=import-outside-toplevel
            from sky.server import plugins as server_plugins
            return server_plugins.get_plugin_rbac_rules()
        except ImportError:
            # Plugin module not available (e.g., not running as server)
            logger.debug(
                'Plugin module not available, skipping plugin RBAC rules')
            return {}
        except Exception as e:  # pylint: disable=broad-except
            logger.warning(f'Failed to get plugin RBAC rules: {e}')
            return {}

    def _maybe_initialize_basic_auth_user(self) -> None:
        """Initialize basic auth user if it is enabled."""
        basic_auth = os.environ.get(constants.SKYPILOT_INITIAL_BASIC_AUTH)
        if not basic_auth:
            return
        username, password = basic_auth.split(':', 1)
        if username and password:
            user_hash = hashlib.md5(
                username.encode()).hexdigest()[:common_utils.USER_HASH_LENGTH]
            user_info = global_user_state.get_user(user_hash)
            if user_info:
                logger.debug(f'Basic auth user {username} already exists')
                return
            global_user_state.add_or_update_user(
                models.User(id=user_hash,
                            name=username,
                            password=password,
                            user_type=models.UserType.BASIC.value))
            enforcer = self._ensure_enforcer()
            enforcer.add_grouping_policy(user_hash, rbac.RoleName.ADMIN.value)
            enforcer.save_policy()
            logger.info(f'Basic auth user {username} initialized')

    def _maybe_initialize_policies(self) -> None:
        """Initialize policies if they don't already exist."""
        logger.debug(f'Initializing policies in process: {os.getpid()}')

        policy_updated = False

        # Check if policies are already initialized by looking for existing
        # permission policies in the enforcer
        enforcer = self._ensure_enforcer()
        # Convert existing policies to set of tuples for O(1) lookups
        existing_policies = {tuple(p) for p in enforcer.get_policy()}

        # Get plugin RBAC rules dynamically
        plugin_rules = self._get_plugin_rbac_rules()

        # If we already have policies for the expected roles, skip
        # initialization
        role_permissions = rbac.get_role_permissions(plugin_rules=plugin_rules)
        expected_policies = []
        for role, permissions in role_permissions.items():
            if permissions.get('permissions'
                              ) and 'blocklist' in permissions['permissions']:
                blocklist = permissions['permissions']['blocklist']
                for item in blocklist:
                    expected_policies.append(
                        (role, item['path'], item['method']))

        # Add workspace policy
        workspace_policy_permissions = rbac.get_workspace_policy_permissions()
        logger.debug(f'Workspace policy permissions from config: '
                     f'{workspace_policy_permissions}')

        for workspace_name, users in workspace_policy_permissions.items():
            for user in users:
                expected_policies.append((user, workspace_name, '*'))
        # Check if all expected policies already exist and find missing ones
        missing_policies = [
            p for p in expected_policies if p not in existing_policies
        ]
        # Find policies to remove
        expected_policies_set = set(expected_policies)
        redundant_policies = [
            p for p in existing_policies if p not in expected_policies_set
        ]
        if missing_policies:
            # Add missing policies
            logger.debug(f'Found {len(missing_policies)} missing policies, '
                         'initializing...')
            for p in missing_policies:
                logger.debug(f'Adding policy: {p}')
                enforcer.add_policy(*p)
                policy_updated = True
            logger.debug('Missing policies added successfully')

        if redundant_policies:
            # Remove redundant policies
            logger.debug(f'Found {len(redundant_policies)} redundant policies, '
                         'cleaning up...')
            for p in redundant_policies:
                logger.debug(f'Removing policy: {p}')
                enforcer.remove_policy(*p)
                policy_updated = True
            logger.debug('Redundant policies removed successfully')

        if not missing_policies and not redundant_policies:
            logger.debug('Policies already in sync, skipping initialization')

        # Always ensure users have default roles (this is idempotent)
        # Get users who already have roles (g policies) to avoid redundant calls
        users_with_roles = {tuple(g)[0] for g in enforcer.get_grouping_policy()}
        all_users = global_user_state.get_all_users()
        for existing_user in all_users:
            if str(existing_user.id) not in users_with_roles:
                logger.debug(f'Adding role for user: {existing_user.name}'
                             f'({existing_user.id})')
                user_added = self._add_user_if_not_exists_no_lock(
                    existing_user.id)
                policy_updated = policy_updated or user_added
        for system_user_id in [
                common.SERVER_ID, constants.SKYPILOT_SYSTEM_USER_ID
        ]:
            if system_user_id not in users_with_roles:
                logger.debug(f'Adding role for system user: {system_user_id}')
                user_added = self._add_user_if_not_exists_no_lock(
                    system_user_id, rbac.RoleName.ADMIN.value)
                policy_updated = policy_updated or user_added
        if policy_updated:
            enforcer.save_policy()

    def add_user_if_not_exists(self, user_id: str) -> None:
        """Add user role relationship."""
        self._lazy_initialize()
        with _policy_lock():
            self._add_user_if_not_exists_no_lock(user_id)

    def _add_user_if_not_exists_no_lock(self,
                                        user_id: str,
                                        role: Optional[str] = None) -> bool:
        """Add user role relationship without lock.

        Returns:
            True if the user was added, False otherwise.
        """
        enforcer = self._ensure_enforcer()
        user_roles = enforcer.get_roles_for_user(user_id)
        if not user_roles:
            enforcer.add_grouping_policy(user_id, role or
                                         rbac.get_default_role())
            return True
        return False

    def delete_user(self, user_id: str) -> None:
        """Delete user role relationship."""
        with _policy_lock():
            self._load_policy_no_lock()
            enforcer = self._ensure_enforcer()
            current_roles = enforcer.get_roles_for_user(user_id)
            if not current_roles:
                logger.debug(f'User {user_id} has no roles')
                return
            enforcer.remove_grouping_policy(user_id, current_roles[0])
            enforcer.save_policy()
            self.invalidate_user_permission_cache(user_id)

    def update_role(self, user_id: str, new_role: str) -> None:
        """Update user role relationship."""
        with _policy_lock():
            self._load_policy_no_lock()
            enforcer = self._ensure_enforcer()
            current_roles = enforcer.get_roles_for_user(user_id)
            if not current_roles:
                logger.debug(f'User {user_id} has no roles')
            else:
                # TODO(hailong): how to handle multiple roles?
                current_role = current_roles[0]
                if current_role == new_role:
                    logger.debug(f'User {user_id} already has role {new_role}')
                    return
                enforcer.remove_grouping_policy(user_id, current_role)

            # Update user role
            enforcer.add_grouping_policy(user_id, new_role)
            enforcer.save_policy()
            # Always invalidate: even a first role assignment can grant
            # workspace access that was previously denied and cached.
            self.invalidate_user_permission_cache(user_id)

    def get_user_roles(self, user_id: str) -> List[str]:
        """Get all roles for a user.

        This method returns all roles that the user has, including inherited
        roles. For example, if a user has role 'admin' and 'admin' inherits
        from 'user', this method will return ['admin', 'user'].

        Args:
            user: The user ID to get roles for.

        Returns:
            A list of role names that the user has.
        """
        self._load_policy_no_lock()
        enforcer = self._ensure_enforcer()
        return enforcer.get_roles_for_user(user_id)

    def get_users_for_role(self, role: str) -> List[str]:
        """Get all users for a role."""
        self._load_policy_no_lock()
        enforcer = self._ensure_enforcer()
        return enforcer.get_users_for_role(role)

    def get_accessible_workspace_names(self, user_id: str,
                                       workspace_names: Set[str]) -> Set[str]:
        """Return workspace names the user can access (batch, O(1) enforcer).

        Use instead of check_workspace_permission in a loop when filtering
        many workspaces, to avoid N enforcer calls.
        """
        if os.getenv(constants.ENV_VAR_IS_SKYPILOT_SERVER) is None:
            return workspace_names
        roles = self.get_user_roles(user_id)
        if rbac.RoleName.ADMIN.value in roles:
            return workspace_names
        enforcer = self._ensure_enforcer()
        # Scan policy rules directly for workspace access.
        # NOTE: this only matches direct (user_id, workspace, '*') and wildcard
        # ('*', workspace, '*') policies.  It does NOT traverse casbin role
        # hierarchies (the g() function in the model matcher).  If role-based
        # workspace grants are ever added, this method must be updated to use
        # enforcer.enforce() per workspace or expand roles via
        # enforcer.get_implicit_permissions_for_user().
        accessible = set()
        for rule in enforcer.get_policy():
            if len(rule) >= 3 and rule[2] == '*' and (rule[0] == user_id or
                                                      rule[0] == '*'):
                if rule[1] in workspace_names:
                    accessible.add(rule[1])
        return accessible

    def check_endpoint_permission(self, user_id: str, path: str,
                                  method: str) -> bool:
        """Check permission."""
        # We intentionally don't load the policy here, as it is a hot path, and
        # we don't support updating the policy.
        # We don't hold the lock for checking permission, as it is read only and
        # it is a hot path in every request. It is ok to have a stale policy,
        # as long as it is eventually consistent.
        # self._load_policy_no_lock()
        enforcer = self._ensure_enforcer()
        return enforcer.enforce(user_id, path, method)

    def _load_policy_no_lock(self):
        """Load policy from storage."""
        enforcer = self._ensure_enforcer()
        enforcer.load_policy()

    def load_policy(self):
        """Load policy from storage with lock."""
        with _policy_lock():
            self._load_policy_no_lock()

    def _workspace_perm_cache_key(self, workspace_name: str,
                                  user_id: str) -> str:
        """Build a KV cache key for a workspace permission entry."""
        return (f'{_WORKSPACE_PERM_CACHE_PREFIX}'
                f'{workspace_name}'
                f'{_WORKSPACE_PERM_CACHE_KEY_SEP}'
                f'{user_id}')

    def invalidate_workspace_permission_cache(self,
                                              workspace_name: str) -> None:
        """Invalidate all cached permission entries for a workspace."""
        prefix = (f'{_WORKSPACE_PERM_CACHE_PREFIX}'
                  f'{workspace_name}'
                  f'{_WORKSPACE_PERM_CACHE_KEY_SEP}')
        kv_cache.delete_cache_entries_by_prefix(prefix)

    def invalidate_user_permission_cache(self, user_id: str) -> None:
        """Invalidate all cached permission entries for a user."""
        kv_cache.delete_cache_entries_by_prefix_suffix(
            prefix=_WORKSPACE_PERM_CACHE_PREFIX,
            suffix=f'{_WORKSPACE_PERM_CACHE_KEY_SEP}{user_id}')

    def check_workspace_permission(self, user_id: str,
                                   workspace_name: str) -> bool:
        """Check workspace permission.

        This method checks if a user has permission to access a specific
        workspace.  Results are cached in a DB-backed KV cache so that all
        server/executor processes share the same view.

        For private workspaces, the user must have explicit permission.

        For public workspaces, the permission is granted via a wildcard policy
        ('*').
        """
        if os.getenv(constants.ENV_VAR_IS_SKYPILOT_SERVER) is None:
            # When it is not on API server, we allow all users to access all
            # workspaces, as the workspace check has been done on API server.
            return True

        # Check DB-backed KV cache (covers both admin and non-admin results).
        cache_key = self._workspace_perm_cache_key(workspace_name, user_id)
        cached = kv_cache.get_cache_entry(cache_key)
        if cached is not None:
            return cached == '1'

        # Cache miss — compute the permission.
        # Admin users have access to all workspaces.
        role = self.get_user_roles(user_id)
        if rbac.RoleName.ADMIN.value in role:
            result = True
        else:
            # The Casbin model matcher already handles the wildcard '*' case:
            # m = (g(r.sub, p.sub)|| p.sub == '*') && r.obj == p.obj &&
            # r.act == p.act
            # This means if there's a policy ('*', workspace_name, '*'), it
            # will match any user
            enforcer = self._ensure_enforcer()
            result = enforcer.enforce(user_id, workspace_name, '*')

        logger.debug(f'Workspace permission check: user={user_id}, '
                     f'workspace={workspace_name}, result={result}')

        # Cache the result; failures are non-critical.
        try:
            kv_cache.add_or_update_cache_entry(
                cache_key, '1' if result else '0',
                time.time() + _WORKSPACE_PERM_CACHE_TTL_SECONDS)
        except Exception as e:  # pylint: disable=broad-except
            logger.debug(f'Failed to cache workspace permission: {e}')

        return result

    def check_service_account_token_permission(self, user_id: str,
                                               token_owner_id: str,
                                               action: str) -> bool:
        """Check service account token permission.

        This method checks if a user has permission to perform an action on
        a service account token owned by another user.

        Args:
            user_id: The ID of the user requesting the action
            token_owner_id: The ID of the user who owns the token
            action: The action being performed (e.g., 'delete', 'view')

        Returns:
            True if the user has permission, False otherwise
        """
        del action
        # Users can always manage their own tokens
        if user_id == token_owner_id:
            return True

        # Check if user has admin role (admins can manage any token)
        user_roles = self.get_user_roles(user_id)
        if rbac.RoleName.ADMIN.value in user_roles:
            return True

        # Regular users cannot manage tokens owned by others
        return False

    def add_workspace_policy(self, workspace_name: str,
                             users: List[str]) -> None:
        """Add workspace policy.

        Args:
            workspace_name: Name of the workspace
            users: List of user IDs that should have access.
                   For public workspaces, this should be ['*'].
                   For private workspaces, this should be specific user IDs.
        """
        with _policy_lock():
            enforcer = self._ensure_enforcer()
            for user in users:
                logger.debug(f'Adding workspace policy: user={user}, '
                             f'workspace={workspace_name}')
                enforcer.add_policy(user, workspace_name, '*')
            enforcer.save_policy()
            # Invalidate stale cached denials (e.g. from checks between a
            # workspace deletion and its re-creation with the same name).
            self.invalidate_workspace_permission_cache(workspace_name)

    def update_workspace_policy(self, workspace_name: str,
                                users: List[str]) -> None:
        """Update workspace policy.

        Args:
            workspace_name: Name of the workspace
            users: List of user IDs that should have access.
                   For public workspaces, this should be ['*'].
                   For private workspaces, this should be specific user IDs.
        """
        with _policy_lock():
            self._load_policy_no_lock()
            enforcer = self._ensure_enforcer()
            # Remove all existing policies for this workspace
            enforcer.remove_filtered_policy(1, workspace_name)
            # Add new policies
            for user in users:
                logger.debug(f'Updating workspace policy: user={user}, '
                             f'workspace={workspace_name}')
                enforcer.add_policy(user, workspace_name, '*')
            enforcer.save_policy()
            # Invalidate cached permission entries after the policy is
            # persisted so other processes re-compute permissions on next
            # check.
            self.invalidate_workspace_permission_cache(workspace_name)

    def remove_workspace_policy(self, workspace_name: str) -> None:
        """Remove workspace policy."""
        with _policy_lock():
            enforcer = self._ensure_enforcer()
            enforcer.remove_filtered_policy(1, workspace_name)
            enforcer.save_policy()
            # Invalidate cached permission entries after the policy is
            # persisted so other processes re-compute permissions on next
            # check.
            self.invalidate_workspace_permission_cache(workspace_name)


@contextlib.contextmanager
def _policy_lock() -> Generator[None, None, None]:
    """Context manager for policy update lock."""
    try:
        with filelock.FileLock(POLICY_UPDATE_LOCK_PATH,
                               POLICY_UPDATE_LOCK_TIMEOUT_SECONDS):
            yield
    except filelock.Timeout as e:
        raise RuntimeError(f'Failed to reload policy due to a timeout '
                           f'when trying to acquire the lock at '
                           f'{POLICY_UPDATE_LOCK_PATH}. '
                           'Please try again or manually remove the lock '
                           f'file if you believe it is stale.') from e


# Singleton instance of PermissionService for other modules to use.
permission_service = PermissionService()



"""A script that fetches Verda Cloud instance types and pricing info.

This script takes about 1 minute to finish.
"""

import csv
import json
import logging
import os
import re
import sys
from typing import Dict, List, Optional, Tuple

import requests

from sky import sky_logging

logger = sky_logging.init_logger('fetch_verda')


def _get_oauth_token(base_url: str, client_id: str, client_secret: str) -> str:
    """Get OAuth access token using client credentials.

    Args:
        base_url: Base URL for the API
        client_id: Client ID for authentication
        client_secret: Client secret for authentication

    Returns:
        str: Access token
    """

    token_url = f'{base_url}/oauth2/token'
    payload = {
        'grant_type': 'client_credentials',
        'client_id': client_id,
        'client_secret': client_secret,
    }
    headers = {'Content-type': 'application/json'}

    response = requests.post(token_url,
                             json=payload,
                             headers=headers,
                             timeout=30)
    response.raise_for_status()

    token_data = response.json()
    return token_data['access_token']


def _fetch_instance_types(base_url: str, token: str) -> List[Dict]:
    """Fetch all instance types from the API.

    Args:
        base_url: Base URL for the API
        token: OAuth access token

    Returns:
        List[Dict]: List of instance type dictionaries
    """
    url = f'{base_url}/instance-types'
    headers = {'Authorization': f'Bearer {token}'}

    response = requests.get(url, headers=headers, timeout=30)
    response.raise_for_status()

    return response.json()


def _fetch_instance_availability(base_url: str,
                                 token: str,
                                 is_spot: bool = False) -> List[Dict]:
    """Fetch instance availability for different regions.

    Args:
        base_url: Base URL for the API
        token: OAuth access token
        is_spot: Whether to fetch spot availability (True) or on-demand (False)

    Returns:
        List[Dict]: Availability with location_code and availabilities
    """
    url = f'{base_url}/instance-availability'
    headers = {'Authorization': f'Bearer {token}'}
    params = {'is_spot': 'true' if is_spot else 'false'}

    response = requests.get(url, headers=headers, params=params, timeout=30)
    response.raise_for_status()

    return response.json()


def _build_availability_map(
        base_url: str, token: str) -> Dict[str, Dict[str, Tuple[bool, bool]]]:
    """Build a map of instance types to regions and their availability."""
    availability_map: Dict[str, Dict[str, Tuple[bool, bool]]] = {}

    # Fetch on-demand availability
    on_demand_availability = _fetch_instance_availability(base_url,
                                                          token,
                                                          is_spot=False)
    for location_data in on_demand_availability:
        location_code = location_data.get('location_code', '')
        availabilities = location_data.get('availabilities', [])
        for instance_type in availabilities:
            if instance_type not in availability_map:
                availability_map[instance_type] = {}
            if location_code not in availability_map[instance_type]:
                availability_map[instance_type][location_code] = (False, False)
            availability_map[instance_type][location_code] = (
                True,
                availability_map[instance_type][location_code][1],
            )

    # Fetch spot availability
    spot_availability = _fetch_instance_availability(base_url,
                                                     token,
                                                     is_spot=True)
    for location_data in spot_availability:
        location_code = location_data.get('location_code', '')
        availabilities = location_data.get('availabilities', [])
        for instance_type in availabilities:
            if instance_type not in availability_map:
                availability_map[instance_type] = {}
            if location_code not in availability_map[instance_type]:
                availability_map[instance_type][location_code] = (False, False)
            availability_map[instance_type][location_code] = (
                availability_map[instance_type][location_code][0],
                True,
            )

    return availability_map


def _extract_gpu_model(instance: Dict) -> str:
    """Extract GPU model from instance attributes or description.

    Args:
        instance: Instance type dictionary from Verda API

    Returns:
        str: GPU model name or empty string
    """
    # Try to get model from instance dictionary
    if 'model' in instance and instance['model']:
        return instance['model']

    # Fall back to extracting from GPU description
    if ('gpu' in instance and instance['gpu'] and
            'description' in instance['gpu']):
        gpu_desc = instance['gpu'].get('description', '')
        # Extract model name (e.g., "RTX A6000" from "1x NVIDIA RTX A6000 48GB")
        match = re.search(r'(?:NVIDIA\s+)?([A-Z0-9]+\s+[A-Z0-9]+|[A-Z0-9]+)',
                          gpu_desc)
        if match:
            return match.group(1).strip()

    return ''


def _format_accelerator_name(gpu_model: str, gpu_memory_gb: float) -> str:
    """Format accelerator name for catalog format.

    Only A100 models need memory suffix (40GB/80GB) since there are two types.
    Other GPU models should not include memory in the name.
    Also drop Tesla- prefix for Tesla GPUs,
    as SkyPilot assumes all Tesla GPUs are named as "V100"
    """
    if not gpu_model:
        return ''

    # Normalize GPU model name for catalog format
    accelerator_name = gpu_model.replace(' ', '-')

    # No need of "Tesla-" prefix for Tesla GPUs,
    # as SkyPilot assumes all Tesla GPUs are named as "V100"
    if accelerator_name.startswith('Tesla-'):
        accelerator_name = accelerator_name.replace('Tesla-', '')

    # Only A100 needs suffix if it 80GB, or else it is assumed it is 40GB
    if accelerator_name.startswith('A100-'):
        # Check if memory is already in the name
        if gpu_memory_gb == 40:
            accelerator_name = 'A100'
        elif gpu_memory_gb == 80:
            accelerator_name = 'A100-80GB'
        else:
            raise ValueError(f'Unsupported A100 memory: {gpu_memory_gb}')

    if gpu_model != accelerator_name:
        print(f'Accelerator name: {gpu_model} -> {accelerator_name}')
    return accelerator_name


def _build_gpu_info(accelerator_name: str, num_gpus: int,
                    gpu_memory_gb: float) -> str:
    """Build GpuInfo JSON string.

    Args:
        accelerator_name: Formatted accelerator name
        num_gpus: Number of GPUs
        gpu_memory_gb: GPU memory per GPU in GB

    Returns:
        str: JSON string for GpuInfo field
    """
    if num_gpus <= 0 or not accelerator_name or gpu_memory_gb <= 0:
        return ''

    total_gpu_memory_mib = gpu_memory_gb * 1024
    gpu_memory_mib = int(total_gpu_memory_mib / num_gpus)
    gpu_info_dict = {
        'Gpus': [{
            'Name': accelerator_name,
            'Count': num_gpus,
            'MemoryInfo': {
                'SizeInMiB': gpu_memory_mib
            },
        }],
        'TotalGpuMemoryInMiB': total_gpu_memory_mib,
    }
    # Convert to JSON string (csv.writer will handle proper escaping)
    return json.dumps(gpu_info_dict)


def create_catalog(output_path: str) -> None:
    """Create Verda Cloud catalog by fetching data from API.

    Args:
        output_path: Path to output CSV file
    """

    # Get authentication credentials
    client_id: Optional[str] = None
    client_secret: Optional[str] = None
    base_url: Optional[str] = None

    config_file_path = os.path.expanduser('~/.verda/config.json')
    if os.path.exists(config_file_path):
        with open(config_file_path, 'r', encoding='utf-8') as f:
            config = json.load(f)
            client_id = config.get('client_id', client_id)
            client_secret = config.get('client_secret', client_secret)
            base_url = config.get('base_url', base_url)

    client_id = os.environ.get('VERDA_CLIENT_ID', client_id)
    client_secret = os.environ.get('VERDA_CLIENT_SECRET', client_secret)
    base_url = os.environ.get('VERDA_BASE_URL', base_url)

    if not base_url:
        base_url = 'https://api.verda.com/v1'

    if not client_id or not client_secret:
        raise Exception('Verda Cloud configuration not found. '
                        'Please set VERDA_CLIENT_ID and VERDA_CLIENT_SECRET '
                        'environment variables or create ~/.verda/config.json.')

    # Get OAuth token
    logger.info('Authenticating with Verda Cloud API...')
    token = _get_oauth_token(base_url, client_id, client_secret)

    # Fetch instance types
    logger.info('Fetching instance types...')
    instance_types = _fetch_instance_types(base_url, token)
    logger.info(f'Fetched {len(instance_types)} instance types')

    # Fetch availability information
    logger.info('Fetching instance availability...')
    availability_map = _build_availability_map(base_url, token)
    logger.info(
        f'Fetched availability for {len(availability_map)} instance types')

    # Create output directory if needed
    os.makedirs(os.path.dirname(output_path), exist_ok=True)

    # Create CSV file
    logger.info(f'Writing catalog to {output_path}')
    with open(output_path, 'w', encoding='utf-8') as f:
        writer = csv.writer(f, quoting=csv.QUOTE_MINIMAL)

        # Write header
        writer.writerow([
            'InstanceType',
            'UpstreamCloudId',
            'vCPUs',
            'MemoryGiB',
            'AcceleratorName',
            'AcceleratorCount',
            'GpuInfo',
            'Region',
            'Price',
            'SpotPrice',
        ])

        for instance in instance_types:
            try:
                # Extract data from instance dictionary
                instance_type_id = instance.get('instance_type', '')
                cpu_data = instance.get('cpu', {})
                memory_data = instance.get('memory', {})
                gpu_data = instance.get('gpu', {})
                gpu_memory_data = instance.get('gpu_memory', {})

                vcpus = (float(cpu_data.get('number_of_cores', 0))
                         if cpu_data else 0)
                memory_gib = (float(memory_data.get('size_in_gigabytes', 0))
                              if memory_data else 0)
                price = float(instance.get('price_per_hour', 0))
                # Get spot price if available, otherwise use regular price
                spot_price_raw = instance.get('spot_price') or instance.get(
                    'spot_price_per_hour')
                spot_price = float(spot_price_raw) if spot_price_raw else price

                # GPU information
                num_gpus = gpu_data.get('number_of_gpus', 0) if gpu_data else 0
                gpu_memory_gb = (gpu_memory_data.get('size_in_gigabytes', 0)
                                 if gpu_memory_data else 0)

                # Extract and format GPU model
                if num_gpus > 0:
                    gpu_model = _extract_gpu_model(instance)
                    accelerator_name = _format_accelerator_name(
                        gpu_model, int(gpu_memory_gb / num_gpus))
                    gpu_info = _build_gpu_info(accelerator_name, num_gpus,
                                               gpu_memory_gb)
                else:
                    accelerator_name = ''
                    gpu_info = ''

                # Get available regions for this instance type
                available_regions = availability_map.get(instance_type_id, {})

                # Only include regions reported by the availability API
                if not available_regions:
                    continue

                # Write row(s) for each available region
                for region, availability_tuple in available_regions.items():
                    on_demand_available, spot_available = availability_tuple

                    effective_price = price if on_demand_available else ''
                    effective_spot_price = (spot_price
                                            if spot_available else '')

                    writer.writerow([
                        instance_type_id,
                        instance_type_id,
                        vcpus,
                        memory_gib,
                        accelerator_name,
                        float(num_gpus) if num_gpus > 0 else '',
                        gpu_info,
                        region,
                        effective_price,
                        effective_spot_price,
                    ])
            except Exception as e:  # pylint: disable=broad-except
                instance_type_id = instance.get('instance_type', 'unknown')
                logger.warning(
                    f'Error processing instance type {instance_type_id}: {e}')
                continue

    logger.info(f'Verda catalog saved to {output_path}')
    logger.info(f'Processed {len(instance_types)} instance types')


def main() -> None:
    """Write SkyPilot v8 catalog for Verda Cloud.

    This is authenticated API request, so you need Verda Cloud API
    credentials configured, in ~/.verda/config.json or in environment variables
    VERDA_CLIENT_ID and VERDA_CLIENT_SECRET.

    Will write to ~/.sky/catalogs/v8/verda/vms.csv if no file name is provided.

    > python sky/catalog/data_fetchers/fetch_verda.py [<file-name>]
    """
    logging.basicConfig(level=logging.INFO,)
    args = sys.argv[1:]
    output_path = args[0] if args else '~/.sky/catalogs/v8/verda/vms.csv'
    output_file = os.path.expanduser(output_path)
    create_catalog(output_file)
    logger.info('Done!')


if __name__ == '__main__':
    main()



"""Scheduler for managed jobs.

Once managed jobs are submitted via submit_job, the scheduler is responsible for
the business logic of deciding when they are allowed to start, and choosing the
right one to start. The scheduler will also schedule jobs that are already live
but waiting to launch a new task or recover.

The scheduler is not its own process - instead, maybe_schedule_next_jobs() can
be called from any code running on the managed jobs controller instance to
trigger scheduling of new jobs if possible. This function should be called
immediately after any state change that could result in jobs newly being able to
be scheduled. If the job is running in a pool, the scheduler will only schedule
jobs for the same pool, because the resources limitations are per-pool (see the
following section for more details).

The scheduling logic limits #running jobs according to three limits:
1. The number of jobs that can be launching (that is, STARTING or RECOVERING) at
   once, based on the number of CPUs. This the most compute-intensive part of
   the job lifecycle, which is why we have an additional limit.
   See sky/utils/controller_utils.py::_get_launch_parallelism.
2. The number of jobs that can be running at any given time, based on the amount
   of memory. Since the job controller is doing very little once a job starts
   (just checking its status periodically), the most significant resource it
   consumes is memory.
   See sky/utils/controller_utils.py::_get_job_parallelism.
3. The number of jobs that can be running in a pool at any given time, based on
   the number of ready workers in the pool. (See _can_start_new_job.)

The state of the scheduler is entirely determined by the schedule_state column
of all the jobs in the job_info table. This column should only be modified via
the functions defined in this file. We will always hold the lock while modifying
this state. See state.ManagedJobScheduleState.

Nomenclature:
- job: same as managed job (may include multiple tasks)
- launch/launching: launching a cluster (sky.launch) as part of a job
- start/run: create the job controller process for a job
- schedule: transition a job to the LAUNCHING state, whether a new job or a job
  that is already alive
- alive: a job controller exists (includes multiple schedule_states: ALIVE,
  ALIVE_WAITING, LAUNCHING)
"""

from argparse import ArgumentParser
import asyncio
import contextlib
import os
import pathlib
import shutil
import sys
import typing
from typing import List, Optional, Set
import uuid

import filelock

from sky import sky_logging
from sky import skypilot_config
from sky.adaptors import common as adaptors_common
from sky.client import sdk
from sky.jobs import constants as managed_job_constants
from sky.jobs import file_content_utils
from sky.jobs import state
from sky.jobs import utils as managed_job_utils
from sky.skylet import constants
from sky.utils import controller_utils
from sky.utils import dag_utils
from sky.utils import subprocess_utils

if typing.TYPE_CHECKING:
    import logging

    import psutil
else:
    psutil = adaptors_common.LazyImport('psutil')

logger = sky_logging.init_logger('sky.jobs.controller')

# Job controller lock. This is used to synchronize writing/reading the
# controller pid file.
JOB_CONTROLLER_PID_LOCK = os.path.expanduser(
    '~/.sky/locks/job_controller_pid.lock')

JOB_CONTROLLER_PID_PATH = os.path.expanduser('~/.sky/job_controller_pid')
JOB_CONTROLLER_ENV_PATH = os.path.expanduser('~/.sky/job_controller_env')

CURRENT_HASH = os.path.expanduser('~/.sky/wheels/current_sky_wheel_hash')


def _parse_controller_pid_entry(
        entry: str) -> Optional[state.ControllerPidRecord]:
    entry = entry.strip()
    if not entry:
        return None
    # The entry should be like <pid>,<started_at>
    # pid is an integer, started_at is a float
    # For backwards compatibility, we also support just <pid>
    entry_parts = entry.split(',')
    if len(entry_parts) == 2:
        [raw_pid, raw_started_at] = entry_parts
    elif len(entry_parts) == 1:
        # Backwards compatibility, pre-#7847
        # TODO(cooperc): Remove for 0.13.0
        raw_pid = entry_parts[0]
        raw_started_at = None
    else:
        # Unknown format
        return None

    try:
        pid = int(raw_pid)
    except ValueError:
        return None

    started_at: Optional[float] = None
    if raw_started_at:
        try:
            started_at = float(raw_started_at)
        except ValueError:
            started_at = None
    return state.ControllerPidRecord(pid=pid, started_at=started_at)


def get_controller_process_records(
) -> Optional[List[state.ControllerPidRecord]]:
    """Return recorded controller processes if the file can be read."""
    if not os.path.exists(JOB_CONTROLLER_PID_PATH):
        # If the file doesn't exist, it means the controller server is not
        # running, so we return an empty list
        return []
    try:
        with open(JOB_CONTROLLER_PID_PATH, 'r', encoding='utf-8') as f:
            lines = f.read().splitlines()
    except (FileNotFoundError, OSError):
        return None

    records: List[state.ControllerPidRecord] = []
    for line in lines:
        record = _parse_controller_pid_entry(line)
        if record is not None:
            records.append(record)
    return records


def _append_controller_pid_record(pid: int,
                                  started_at: Optional[float]) -> None:
    # Note: started_at is a float, but converting to a string will not lose any
    # precision. See https://docs.python.org/3/tutorial/floatingpoint.html and
    # https://github.com/python/cpython/issues/53583
    entry = str(pid) if started_at is None else f'{pid},{started_at}'
    with open(JOB_CONTROLLER_PID_PATH, 'a', encoding='utf-8') as f:
        f.write(entry + '\n')


def start_controller() -> None:
    """Start the job controller process.

    This requires that the env file is already set up.
    """
    os.environ[constants.OVERRIDE_CONSOLIDATION_MODE] = 'true'
    logs_dir = os.path.expanduser(
        managed_job_constants.JOBS_CONTROLLER_LOGS_DIR)
    os.makedirs(logs_dir, exist_ok=True)
    controller_uuid = str(uuid.uuid4())
    log_path = os.path.join(logs_dir, f'controller_{controller_uuid}.log')

    activate_python_env_cmd = (f'{constants.ACTIVATE_SKY_REMOTE_PYTHON_ENV};')
    run_controller_cmd = (f'{sys.executable} -u -m'
                          f'sky.jobs.controller {controller_uuid}')

    run_cmd = (f'{activate_python_env_cmd}'
               f'{run_controller_cmd}')

    logger.info(f'Running controller with command: {run_cmd}')

    pid = subprocess_utils.launch_new_process_tree(run_cmd, log_output=log_path)
    pid_started_at = psutil.Process(pid).create_time()
    _append_controller_pid_record(pid, pid_started_at)


def get_alive_controllers() -> Optional[int]:
    records = get_controller_process_records()
    if records is None:
        # If we cannot read the file reliably, avoid starting extra controllers.
        return None
    if not records:
        return 0

    alive = 0
    for record in records:
        if managed_job_utils.controller_process_alive(record, quiet=False):
            alive += 1
    return alive


def maybe_start_controllers(from_scheduler: bool = False) -> None:
    """Start the job controller process.

    If the process is already running, it will not start a new one.
    Will also add the job_id, dag_yaml_path, and env_file_path to the
    controllers list of processes.
    """
    # In consolidation mode, during rolling update, two API servers may be
    # running. If we are on the new API server, and we haven't finished the
    # recovery process, we should avoid starting new controllers. The old API
    # server/consolidated jobs controller could run update_managed_jobs_statuses
    # and if there are jobs running on the new API server, the old one will not
    # see the corresponding processes and may mark them as FAILED_CONTROLLER.
    if from_scheduler and managed_job_utils.is_consolidation_mode(
    ) and os.path.exists(
            os.path.expanduser(
                constants.PERSISTENT_RUN_RESTARTING_SIGNAL_FILE)):
        # This could happen during an API server rolling update, or during
        # normal running while managed-job-status-refresh-daemon is running. In
        # either case, the controllers should be already started or will be
        # started by the recovery process.
        logger.info('Recovery is still in progress, skipping controller start.')
        return
    try:
        with filelock.FileLock(JOB_CONTROLLER_PID_LOCK, blocking=False):
            if from_scheduler and not managed_job_utils.is_consolidation_mode():
                cur = pathlib.Path(CURRENT_HASH)
                old = pathlib.Path(f'{CURRENT_HASH}.old')

                if old.exists() and cur.exists():
                    if (old.read_text(encoding='utf-8') !=
                            cur.read_text(encoding='utf-8')):
                        # TODO(luca): there is a 1/2^160 chance that there will
                        # be a collision. using a geometric distribution and
                        # assuming one update a day, we expect a bug slightly
                        # before the heat death of the universe. should get
                        # this fixed before then.
                        try:
                            # this will stop all the controllers and the api
                            # server.
                            sdk.api_stop()
                            # All controllers should be dead. Remove the PIDs so
                            # that update_managed_jobs_statuses won't think they
                            # have failed.
                            state.reset_jobs_for_recovery()
                        except Exception as e:  # pylint: disable=broad-except
                            logger.error(f'Failed to stop the api server: {e}')
                            pass
                        else:
                            shutil.copyfile(cur, old)
                if not old.exists():
                    shutil.copyfile(cur, old)

            alive = get_alive_controllers()
            if alive is None:
                return
            wanted = controller_utils.get_number_of_jobs_controllers()
            started = 0

            while alive + started < wanted:
                start_controller()
                started += 1

            if started > 0:
                logger.info(f'Started {started} controllers')

    except filelock.Timeout:
        # If we can't get the lock, just exit. The process holding the lock
        # should launch any pending jobs.
        pass


def submit_jobs(job_ids: List[int],
                dag_yaml_path: str,
                original_user_yaml_path: str,
                env_file_path: str,
                priority: int,
                priority_class: Optional[str] = None) -> None:
    """Submit multiple existing jobs to the scheduler.

    This should be called after jobs are created in the `spot` table as
    PENDING. It will tell the scheduler to try and start the job controllers, if
    there are resources available.

    The user hash should be set (e.g. via SKYPILOT_USER_ID) before calling this.
    """
    job_ids_without_controller_process = []
    for job_id in job_ids:
        controller_process = state.get_job_controller_process(job_id)
        if controller_process is not None:
            # why? TODO(cooperc): figure out why this is needed, fix it, and
            # remove
            if managed_job_utils.controller_process_alive(
                    controller_process, job_id):
                # This can happen when HA recovery runs for some reason but the
                # job controller is still alive.
                logger.warning(f'Job {job_id} is still alive with controller '
                               f'{controller_process}, skipping submission')
                continue
        job_ids_without_controller_process.append(job_id)
    job_ids = job_ids_without_controller_process

    with open(dag_yaml_path, 'r', encoding='utf-8') as dag_file:
        dag_yaml_content = dag_file.read()
    with open(original_user_yaml_path, 'r',
              encoding='utf-8') as original_user_yaml_file:
        original_user_yaml_content = original_user_yaml_file.read()
    with open(env_file_path, 'r', encoding='utf-8') as env_file:
        env_file_content = env_file.read()

    # Read config file if SKYPILOT_CONFIG env var is set
    config_file_content: Optional[str] = None
    config_file_path = os.environ.get(skypilot_config.ENV_VAR_SKYPILOT_CONFIG)
    if config_file_path:
        config_file_path = os.path.expanduser(config_file_path)
        if os.path.exists(config_file_path):
            with open(config_file_path, 'r', encoding='utf-8') as config_file:
                config_file_content = config_file.read()

    config_bytes = (len(config_file_content) if config_file_content else 0)
    logger.debug(f'Storing jobs {job_ids} file contents in database '
                 f'(DAG bytes={len(dag_yaml_content)}, '
                 f'original user yaml bytes={len(original_user_yaml_content)}, '
                 f'env bytes={len(env_file_content)}, '
                 f'config bytes={config_bytes}).')

    # Submit all jobs
    state.scheduler_set_waiting(job_ids, dag_yaml_content,
                                original_user_yaml_content, env_file_content,
                                config_file_content, priority, priority_class)
    maybe_start_controllers(from_scheduler=True)


@contextlib.asynccontextmanager
async def scheduled_launch(
    job_id: int,
    starting: Set[int],
    starting_lock: asyncio.Lock,
    starting_signal: asyncio.Condition,
):
    """Launch as part of an ongoing job.

    A newly started job will already be LAUNCHING, and this will immediately
    enter the context.

    If a job is ongoing (ALIVE schedule_state), there are two scenarios where we
    may need to call sky.launch again during the course of a job controller:
    - for tasks after the first task
    - for recovery

    This function will mark the job as ALIVE_WAITING, which indicates to the
    scheduler that it wants to transition back to LAUNCHING. Then, it will wait
    until the scheduler transitions the job state, before entering the context.

    On exiting the context, the job will transition to ALIVE.

    This should only be used within the job controller for the given job_id. If
    multiple uses of this context are nested, behavior is undefined. Don't do
    that.
    """
    pool = state.get_pool_from_job_id(job_id)
    # For pool, since there is no execution.launch, we don't need to have all
    # the ALIVE_WAITING state. The state transition will be
    # WAITING -> ALIVE -> DONE without any intermediate transitions.
    if pool is not None:
        yield
        return

    # For JobGroups, multiple tasks share the same job_id but each launches
    # a different cluster in parallel. We handle scheduler state at the group
    # level in _run_job_group(), so bypass per-task scheduling here.
    # Check if job is a JobGroup by examining the DAG YAML content.
    # TODO(zhwu): make JobGroup scheduler aware.
    dag_content = file_content_utils.get_job_dag_content(job_id)
    if dag_content is not None and dag_utils.is_job_group_yaml_str(dag_content):
        yield
        return

    assert starting_lock == starting_signal._lock, (  # type: ignore #pylint: disable=protected-access
        'starting_lock and starting_signal must use the same lock')

    while True:
        async with starting_lock:
            starting_count = len(starting)
            if starting_count < controller_utils.LAUNCHES_PER_WORKER:
                break
            logger.info('Too many jobs starting, waiting for a slot')
            await starting_signal.wait()

    logger.info(f'Starting job {job_id}')

    async with starting_lock:
        starting.add(job_id)

    await state.scheduler_set_launching_async(job_id)

    try:
        yield
    except Exception as e:
        raise e
    else:
        await state.scheduler_set_alive_async(job_id)
    finally:
        async with starting_lock:
            starting.remove(job_id)
            starting_signal.notify()


def job_done(job_id: int, idempotent: bool = False) -> None:
    """Transition a job to DONE.

    If idempotent is True, this will not raise an error if the job is already
    DONE.

    The job could be in any terminal ManagedJobStatus. However, once DONE, it
    should never transition back to another state.

    This is only called by utils.update_managed_jobs_statuses which is sync.
    """
    if idempotent and (state.get_job_schedule_state(job_id)
                       == state.ManagedJobScheduleState.DONE):
        return

    state.scheduler_set_done(job_id, idempotent)


async def job_done_async(job_id: int, idempotent: bool = False):
    """Async version of job_done."""
    if idempotent and (await state.get_job_schedule_state_async(job_id)
                       == state.ManagedJobScheduleState.DONE):
        return

    await state.scheduler_set_done_async(job_id, idempotent)


if __name__ == '__main__':
    parser = ArgumentParser()
    parser.add_argument('dag_yaml',
                        type=str,
                        help='The path to the user job yaml file.')
    parser.add_argument('--user-yaml-path',
                        type=str,
                        help='The path to the original user job yaml file.')
    parser.add_argument(
        '--job-id',
        type=int,
        nargs='+',
        help='Job id(s) for the controller job(s). Can specify multiple.')
    parser.add_argument('--env-file',
                        type=str,
                        help='The path to the controller env file.')
    parser.add_argument('--pool',
                        type=str,
                        required=False,
                        default=None,
                        help='The pool to use for the controller job.')
    parser.add_argument(
        '--priority',
        type=int,
        default=constants.DEFAULT_PRIORITY,
        help=
        f'Job priority ({constants.MIN_PRIORITY} to {constants.MAX_PRIORITY}).'
        f' Default: {constants.DEFAULT_PRIORITY}.')
    parser.add_argument('--priority-class',
                        type=str,
                        default=None,
                        help='Named priority class for the job.')
    args = parser.parse_args()

    submit_jobs(args.job_id, args.dag_yaml, args.user_yaml_path, args.env_file,
                args.priority, args.priority_class)
